import sys
import re
from pathlib import Path

import numpy as np
import torch
from torch.utils import data
from tqdm import tqdm

from specollate.src.snapconfig import config
from specollate.src.snapsearch import preprocess
from specollate.src.snaputils import simulatespectra as sim


class PeptideDataset(data.Dataset):
    'Characterizes a dataset for PyTorch'
    def __init__(self, dir_path, decoy=False):
        'Initialization'
        
        in_path = Path(dir_path)
        assert in_path.exists()
        assert in_path.is_dir()

        self.aas            = ['_PAD'] + list(config.AAMass.keys())# + list(config.ModCHAR.values())
        self.aa2idx         = {a:i for i, a in enumerate(self.aas)}
        self.idx2aa         = {i:a for i, a in enumerate(self.aas)}
        
        self.pep_path       = dir_path
        self.vocab_size     = len(self.aa2idx) # + self.charge + self.num_species + 1
        print("Vocabulary size: {}".format(self.vocab_size))
        self.seq_len        = config.get_config(section='ml', key='pep_seq_len')
        
        print("Loading peptides...")
        pep_lst, prot_list, pep_mass_lst, pep_modified_lst = load_peps(self.pep_path)
        
        self.pep_lst_set = set(pep_lst)

        print("peptide list len: {}".format(len(pep_lst)))
        print("peptide set len: {}".format(len(self.pep_lst_set)))
#         out_dir = "/disk/raptor-2/mtari008/data/deepsnap/preprocessed-human-hcd-tryp-best/pts/"
#         with open(join(out_dir, 'pep_pickle.pkl'), 'rb') as f:
#             search_peps = pickle.load(f)
#         added_counter = 0
#         for s_pep in search_peps:
#             if s_pep not in self.pep_lst_set:
#                 self.pep_lst_set.add(s_pep)
#                 added_counter += 1
#                 pep_lst.append(s_pep)
#                 prot_list.append("unknown")
#                 pep_mass_lst.append(sim.get_pep_mass(s_pep))
#                 pep_modified_lst.append(any(aa.islower() for aa in s_pep))
#         print("New peptides added: {}".format(added_counter))
            
            
        all_sorts = list(zip(*sorted(zip(pep_lst, prot_list, pep_mass_lst, pep_modified_lst), key=lambda x: x[2])))
        self.pep_list          = all_sorts[0]
        self.prot_list         = all_sorts[1]
        self.pep_mass_list     = all_sorts[2]
        self.pep_modified_list = all_sorts[3]
        if decoy:
            self.pep_list, self.prot_list, self.pep_mass_list, self.pep_modified_list = self.get_docoys()
            
        print('Peptide Dataset Size: {}'.format(len(self.pep_list)))
        

    def __len__(self):
        'Denotes the total number of samples'
        return len(self.pep_list)


    def __getitem__(self, index):
        'Generates one sample of data'
        pep = self.pep_list[index].strip()
        pepl = [self.aa2idx[aa] for aa in pep]
        pepl = self.pad_left(pepl)

        # pep_mass = sim.get_pep_mass(pep)
        # gray_mass = sim.gray_code(round(pep_mass * 100))
        # mass_arr = sim.decimal_to_binary_array(gray_mass, 24)
        # pepl = np.concatenate((mass_arr, pepl))
        
        torch_pep = torch.tensor(pepl, dtype=torch.long)
        return torch_pep
        

    def pad_left(self, arr):
        out = np.zeros(self.seq_len)
        out[-len(arr):] = arr
        return out
    
    def get_docoys(self):
        decoy_list = []
        decoy_prot_list = []
        decoy_mass_list = []
        decoy_modified_list = []
        for pep, prot, mass, modified in zip(self.pep_list, self.prot_list, self.pep_mass_list, self.pep_modified_list):
            pep_parts = re.findall(r"([A-Z][a-z]?)", pep)
            decoy_pep = pep_parts[0] + "".join(pep_parts[-2:0:-1]) + pep_parts[-1]
            if decoy_pep not in self.pep_lst_set:
                decoy_list.append(decoy_pep)
                decoy_prot_list.append(prot)
                decoy_mass_list.append(mass)
                decoy_modified_list.append(modified)
        return decoy_list, decoy_prot_list, decoy_mass_list, decoy_modified_list


### These are not class functions. Use them normally. ###

def find_occurrences(s, ch):
    return [i for i, letter in enumerate(s) if letter == ch]


def apply_mod(peps, mod):
    result_peps = set()
    pep_seq_len = config.get_config(section="ml", key="pep_seq_len")
    for pep in peps:
        if len(pep) >= pep_seq_len:
            continue
        for mod_aa in mod["aas"]:
            if mod_aa == "nt" and not pep[1].islower():
                result_peps.update([pep[0] + mod["mod_char"] + pep[1:]])
            elif mod_aa == "ct" and not pep[-1].islower():
                result_peps.update([pep + mod["mod_char"]])
            else:
                aa_indices = find_occurrences(pep, mod_aa)
                for index in aa_indices:
                    if index == len(pep) - 1 or not pep[index+1].islower():
                        result_peps.update([pep[:index+1] + mod["mod_char"] + (pep[index+1:] if index < len(pep) else "")])
        
    return result_peps


def add_mods(pep, mods, num_mods):
    mod_peps = set([pep])
    result_peps = set([pep])
    for i in range(num_mods):
        temp_mod_peps = set()
        for mod in mods:
            temp_mod_peps.update(apply_mod(mod_peps, mod))
        mod_peps.update(set(temp_mod_peps))
        result_peps.update(mod_peps)
    
    return result_peps

def load_peps(pep_dir):
    fasta_files = preprocess.verify_in_dir(pep_dir, "fasta")
    
    use_mods = config.get_config(key="use_mods", section="input")
    mods_list = config.Mods
    num_mods = config.get_config(key="num_mods", section="search")
    pep_seq_len = config.get_config(key="pep_seq_len", section="ml")
    
    pep_set = set()
    pep_list = []
    masses = []
    modifieds = []
    prot_list = []
    
    tot_pep_count = 0
    for fasta_file in fasta_files:
        tqdm.write('Reading: {}'.format(fasta_file))
        
        f = open(fasta_file, "r")
        lines = f.readlines()
        f.close()
        
        # print("File length: {}".format(len(lines)))
        peps = []
        temp_prot = ""
        pbar = tqdm(lines, file=sys.stdout)
        pbar.set_description('Loading Peptides...')
        # with progressbar.ProgressBar(max_value=len(lines)) as bar:
        for line in pbar:
            line = line.strip().replace("C", "Cc")
            if line.startswith(">"):
                temp_prot = line[1:].strip()
                continue
            elif any(x in line for x in config.Ignore):
                continue
            peps = add_mods(line, mods_list, num_mods) if use_mods else [line]
            for pep in peps:
                pep = pep.strip()
                mass = sim.get_pep_mass(pep)
                modified = any(aa.islower() for aa in pep)
                if pep not in pep_set and len(pep) <= pep_seq_len:
                    pep_set.add(pep)
                    pep_list.append(pep)
                    masses.append(mass)
                    modifieds.append(modified)
                    prot_list.append(temp_prot)
                    tot_pep_count += 1
            # bar.update(i)
        tqdm.write("Peptides written: {}".format(tot_pep_count))
        
        return pep_list, prot_list, masses, modifieds


import random as rand
#from matplotlib import pyplot as plt
import torch
import torch.nn as nn

from specollate.src.snapconfig import config
from specollate.src.snaptrain import process

rand.seed(37)

train_accuracy = []
train_loss = []
test_accuracy = []
test_loss = []
snp_weight = config.get_config(section="ml", key="snp_weight")
ce_weight = config.get_config(section="ml", key="ce_weight")
mse_weight = config.get_config(section="ml", key="mse_weight")
divider = snp_weight + ce_weight# + mse_weight

# def exit_handler():
#     plt.figure()
#     plt.plot(train_accuracy)
#     plt.plot(test_accuracy)
#     plt.xlabel("epoch")
#     plt.ylabel("accuracy")
#     plt.savefig("accuracy.png", dpi=600)

#     plt.figure()
#     plt.plot(train_loss)
#     plt.plot(test_loss)
#     plt.xlabel("epoch")
#     plt.ylabel("loss")
#     plt.savefig("loss.png", dpi=600)


# atexit.register(exit_handler)


def train(model, device, train_loader, triplet_loss, cross_entropy_loss, mse_loss, optimizer, epoch):
    model.train()
    batch_size = config.get_config(section='ml', key='batch_size')
    h = model.module.init_hidden(batch_size)

    accurate_labels = 0
    accurate_labels_ch = 0
    all_labels = 0
    snp_loss = ce_loss = 0
    l = 0
    # with progressbar.ProgressBar(max_value=len(train_loader)) as p_bar:
    for idx, data in enumerate(train_loader):
        h = tuple([e.data for e in h])
        q_len = len(data[0])
        p_len = len(data[1])
        d_len = len(data[2])
        if p_len > d_len:
            seq_len = config.get_config(section='ml', key='pep_seq_len') # + 111
            zero_pad = torch.zeros(p_len - d_len, seq_len, dtype=torch.long)
            data[2] = torch.cat((data[2], zero_pad), 0)
        data[0] = data[0].to(device) # spectra
        data[1] = data[1].to(device) # peptides
        data[2] = data[2].to(device) # decoys
        data[3] = data[3].to(device) # spectrum charges
        # data[4] = data[4].to(device) # spectrum masses
        # data[5] = data[5].to(device) # peptide masses
        # data[6] = data[6].to(device) # decoy masses
        counts = data[4]
        
        optimizer.zero_grad()
        
        # Q, ch, sm, P, pm, D, dm = model(data[:-1])
        Q, P, D = model(data[:-1])
        # assert len(Q) == len(ch)

        snp_loss, QxPD = snap_loss_2_d(counts, P, Q, D[:d_len], triplet_loss, device)
        
        # ce_loss = cross_entropy_loss(ch, data[3])
        # mse_loss_specs = mse_loss(sm, data[4])
        # mse_loss_peps = mse_loss(pm, data[5])
        # mse_loss_decs = mse_loss(dm[:d_len], data[6])
        # mse_tot_loss = mse_loss_specs + mse_loss_peps + mse_loss_decs
        # mse_tot_loss = mse_tot_loss / 3
        
        # loss = (snp_weight * snp_loss + ce_weight * ce_loss) / divider
        loss = snp_loss
        # print("snp loss: {}, ce loss: {}, mse loss: {}".format(snp_loss, ce_loss, mse_tot_loss))
        
        # with amp.scale_loss(loss, optimizer) as scaled_loss:
        loss.backward()
        
        nn.utils.clip_grad_norm_(model.parameters(), 5)
        
        optimizer.step()

        seq, _ = get_index(counts, q_len)
        seq = torch.LongTensor(seq).to(device)
        # QxP contains the distance of each spectrum from each peptide.

        accurate_labels += torch.sum(QxPD.argmin(1) == seq)
        # accurate_labels_ch += multi_acc(ch, data[3])
        all_labels += len(Q)
        # p_bar.update(idx)
    
    accuracy = 100. * float(accurate_labels) / all_labels
    # accuracy_ch = 100. * float(accurate_labels_ch) / all_labels

    # wandb.log({"train loss": snp_loss, "train accuracy": accuracy}, step=epoch)
    print('Train accuracy:\t{}/{} ({:.3f}%)\t\tLoss: {:.6f}'.format(accurate_labels, all_labels, accuracy, snp_loss))
    # print('Train accuracy charge:\t{}/{} ({:.3f}%)\t\tLoss: {:.6f}'.format(accurate_labels_ch, all_labels, accuracy_ch, ce_loss))
    # print('Train loss mse:\t{:.6f}'.format(mse_tot_loss))
    return loss
    

def test(model, device, test_loader, triplet_loss, cross_entropy_loss, mse_loss, epoch):
    model.eval()
    
    with torch.no_grad():
        accurate_labels = 0
        accurate_labels_ch = 0
        all_labels = 0
        snp_loss = ce_loss = 0

        batch_size = config.get_config(section='ml', key='batch_size')
        h = model.module.init_hidden(batch_size)
        
        for data in test_loader:
            h = tuple([e.data for e in h])
            q_len = len(data[0])
            p_len = len(data[1])
            d_len = len(data[2])
            if p_len > d_len:
                seq_len = config.get_config(section='ml', key='pep_seq_len') # + 111
                zero_pad = torch.zeros(p_len - d_len, seq_len, dtype=torch.long)
                data[2] = torch.cat((data[2], zero_pad))
            data[0] = data[0].to(device) # spectra
            data[1] = data[1].to(device) # peptides
            data[2] = data[2].to(device) # decoys
            data[3] = data[3].to(device) # spectrum charges
            # data[4] = data[4].to(device) # spectrum masses
            # data[5] = data[5].to(device) # peptide masses
            # data[6] = data[6].to(device) # decoy masses
            counts = data[4]
            
            # Q, ch, sm, P, pm, D, dm = model(data[:-1])
            Q, P, D = model(data[:-1])

            snp_loss, QxPD = snap_loss_2_d(counts, P, Q, D[:d_len], triplet_loss, device)
            # ce_loss = cross_entropy_loss(ch, data[3])
            # mse_loss_specs = mse_loss(sm, data[4])
            # mse_loss_peps = mse_loss(pm, data[5])
            # mse_loss_decs = mse_loss(dm[:d_len], data[6])
            # mse_tot_loss = mse_loss_specs + mse_loss_peps + mse_loss_decs
            # mse_tot_loss = mse_tot_loss / 3
            # loss = (snp_weight * snp_loss + ce_weight * ce_loss) / divider
            # loss = snp_loss
            
            seq, _ = get_index(counts, q_len)
            seq = torch.LongTensor(seq).to(device)
            # QxP contains the distance of each spectrum from each peptide.
            accurate_labels += torch.sum(QxPD.argmin(1) == seq)
            # accurate_labels_ch += multi_acc(ch, data[3])
            all_labels = all_labels + len(Q)
                
        accuracy = 100. * float(accurate_labels) / all_labels
        # accuracy_ch = 100. * float(accurate_labels_ch) / all_labels
        
        # wandb.log({"test loss": snp_loss, "test accuracy": accuracy}, step=epoch)
        print('Test accuracy:\t{}/{} ({:.3f}%)\t\tLoss: {:.6f}'.format(accurate_labels, all_labels, accuracy, snp_loss))
        # print('Test accuracy charge:\t{}/{} ({:.3f}%)\t\tLoss: {:.6f}'.format(accurate_labels_ch, all_labels, accuracy_ch, ce_loss))
        # print('Test loss mse:\t{:.6f}'.format(mse_tot_loss))


def get_masks(counts, p_len, q_len):
    rows, cols = get_index(counts, q_len)

    PQ_mask = torch.zeros(p_len, q_len)
    PQ_mask[rows, cols] = 1.
    
    QQ_mask = torch.zeros(q_len, q_len)
    QQ_mask[cols, :] = PQ_mask[rows, :]
    
    return PQ_mask, QQ_mask


def get_index(counts, q_len):
    rows = []
    for i, x in enumerate(counts):
        rows.extend([i]*x)
    cols = range(q_len)
    assert len(rows) == len(cols)
    return rows, cols


def snap_loss(counts, Q, P, triplet_loss, device):
    """Mine the hardest sextuplets.""" 
    QxQ = process.pairwise_distances(Q)     # calculate distance matrix for spectra
    PxP = process.pairwise_distances(P)     # calculate distance matrix for peptides
    QxP = process.pairwise_distances(Q, P)  # calculate distance matrix for spectra-peptides
    
    # Get masks
    p_len, q_len = len(P), len(Q)
    PQ_mask, QQ_mask = get_masks(counts, p_len, q_len)
    PQ_mask, QQ_mask = PQ_mask.to(device), QQ_mask.to(device)

    # Mine hardest positives
    PxQ_max = Q[(torch.transpose(QxP, 0, 1) * PQ_mask).max(1).indices]
    QxP_max = P[(QxP * torch.transpose(PQ_mask, 0, 1)).max(1).indices]

    # Mine hardest negatives
    QQ_neg_mask = 1 - QQ_mask
    QQ_neg_mask_clone = QQ_neg_mask.clone()
    QxQ_masked = QxQ * QQ_neg_mask
    QxQ_masked[QQ_neg_mask_clone < 0.1] = float("inf")
    QxQ_min = Q[QxQ_masked.min(1).indices]              # nearest spectrum for each spectrum

    PxP.fill_diagonal_(float("inf"))
    PxP_min = P[PxP.min(1).indices]                     # nearest peptide for each peptide

    PQ_neg_mask = (1 - PQ_mask)
    PQ_neg_mask_clone = PQ_neg_mask.clone()
    PxQ_masked = torch.transpose(QxP, 0, 1) * PQ_neg_mask
    PxQ_masked[PQ_neg_mask_clone < 0.1] = float("inf")
    PxQ_min = Q[PxQ_masked.min(1).indices]              # nearest spectrum for each peptide

    QP_neg_mask = torch.transpose(PQ_neg_mask, 0, 1)
    QP_neg_mask_clone = QP_neg_mask.clone()
    QxP_masked = QxP * QP_neg_mask
    QxP_masked[QP_neg_mask_clone < 0.1] = float("inf")
    QxP_min = P[QxP_masked.min(1).indices]              # nearest peptide for each spectrum

    # Calculate the loss
    loss  = triplet_loss(Q, QxP_max, QxQ_min)   # spectrum-spectrum negatives
    loss += triplet_loss(Q, QxP_max, QxP_min)   # spectrum-peptide negatives
    loss += triplet_loss(P, PxQ_max, PxP_min)   # peptide-peptide negatives
    loss += triplet_loss(P, PxQ_max, PxQ_min)   # peptide-spectrum negatives
    
    loss = loss / 4

    return loss, QxP


def snap_loss_d(counts, P, Q, D, triplet_loss, device):
    """Mine the hardest sextuplets.""" 
    QxQ = process.pairwise_distances(Q)     # calculate distance matrix for spectra
    PxP = process.pairwise_distances(P)     # calculate distance matrix for peptides
    QxP = process.pairwise_distances(Q, P)  # calculate distance matrix for spectra-peptides
    QxD = process.pairwise_distances(Q, D)
    PxD = process.pairwise_distances(P, D)
    
    # Get masks
    p_len, q_len = len(P), len(Q)
    PQ_mask, QQ_mask = get_masks(counts, p_len, q_len)
    PQ_mask, QQ_mask = PQ_mask.to(device), QQ_mask.to(device)

    # Mine hardest positives
    PxQ_max = Q[(torch.transpose(QxP, 0, 1) * PQ_mask).max(1).indices]
    QxP_max = P[(QxP * torch.transpose(PQ_mask, 0, 1)).max(1).indices]

    # Mine hardest negatives
    QQ_neg_mask = 1 - QQ_mask
    QQ_neg_mask_clone = QQ_neg_mask.clone()
    QxQ_masked = QxQ * QQ_neg_mask
    QxQ_masked[QQ_neg_mask_clone < 0.1] = float("inf")
    QxQ_min = Q[QxQ_masked.min(1).indices]       # nearest spectrum for each spectrum

    PD = torch.cat((P, D))
    PxP.fill_diagonal_(float("inf"))
    PxPD = torch.cat((PxP, PxD), axis=1)
    PxPD_min = PD[PxPD.min(1).indices]           # nearest peptide for each peptide

    PQ_neg_mask = (1 - PQ_mask)
    PQ_neg_mask_clone = PQ_neg_mask.clone()
    PxQ_masked = torch.transpose(QxP, 0, 1) * PQ_neg_mask
    PxQ_masked[PQ_neg_mask_clone < 0.1] = float("inf")
    PxQ_min = Q[PxQ_masked.min(1).indices]       # nearest spectrum for each peptide

    QP_neg_mask = torch.transpose(PQ_neg_mask, 0, 1)
    QP_neg_mask_clone = QP_neg_mask.clone()
    QxP_masked = QxP * QP_neg_mask
    QxP_masked[QP_neg_mask_clone < 0.1] = float("inf")
    QxPD = torch.cat((QxP_masked, QxD), axis=1)
    QxPD_min = PD[QxPD.min(1).indices]            # nearest peptide for each spectrum

    # Calculate the loss
    loss  = triplet_loss(Q, QxP_max, QxQ_min)    # spectrum-spectrum negatives
    loss += triplet_loss(Q, QxP_max, QxPD_min)   # spectrum-peptide negatives
    loss += triplet_loss(P, PxQ_max, PxPD_min)   # peptide-peptide negatives
    loss += triplet_loss(P, PxQ_max, PxQ_min)    # peptide-spectrum negatives
    
    loss = loss / 4
    
    QxPD = process.pairwise_distances(Q, PD)
    qpd = torch.cat((QxP, QxD), axis=1)
    return loss, QxPD


def snap_loss_2(counts, P, Q, triplet_loss, device):
    """Mine the hardest sextuplets."""
    PxP = process.pairwise_distances(P)     # calculate distance matrix for peptides
    PxQ = process.pairwise_distances(P, Q)  # calculate distance matrix for spectra-peptides
    
    # # Get masks
    p_len, q_len = len(P), len(Q)
    PQ_mask, _ = get_masks(counts, p_len, q_len)
    PQ_mask = PQ_mask.to(device)

    # Mine hardest positives
    PxQ_masked = PxQ * PQ_mask
    PxQ_max = Q[PxQ_masked.max(1).indices]

    delta_plus = PxQ_masked.max(1).values #this one

    # # Mine hardest negatives
    PxP.fill_diagonal_(float("inf"))
    PxP_min = P[PxP.min(1).indices]              # nearest peptide for each peptide

    delta_minus_P = PxP.min(1).values # this one

    PQ_neg_mask = (1 - PQ_mask).to(device)
    PxQ_masked = PxQ * PQ_neg_mask
    PxQ_masked[PQ_neg_mask < 0.1] = float("inf")
    PxQ_min = Q[PxQ_masked.min(1).indices]          # nearest spectrum for each peptide
    
    delta_minus_Q = PxQ_masked.min(1).values # this one

    margin = config.get_config(section="ml", key="margin")
    filt1 = ((delta_plus - delta_minus_P)) > 0
    # if len(delta_plus) > torch.sum(filt1):
    #     print("filt1 length: {}".format(torch.sum(filt1).item()))
    A1 = P#[filt1]
    P1 = PxQ_max#[filt1]
    N1 = PxP_min#[filt1]

    filt2 = ((delta_plus - delta_minus_Q)) > 0
    # if len(delta_plus) > torch.sum(filt2):
    #     print("filt2 length: {}".format(torch.sum(filt2).item()))
    #     print("out of {}".format(len(delta_plus)))
    A2 = P#[filt2]
    P2 = PxQ_max#[filt2]
    N2 = PxQ_min#[filt2]

    # # Calculate the loss
    loss =  triplet_loss(A1, P1, N1)       # peptide-peptide negatives
    loss += triplet_loss(A2, P2, N2)       # peptide-spectrum negatives  
    loss = loss / 2

    return loss, process.pairwise_distances(Q, P)


def snap_loss_2_d(counts, P, Q, D, triplet_loss, device):
    """Mine the hardest sextuplets."""
    PxP = process.pairwise_distances(P)     # calculate distance matrix for peptides
    PxQ = process.pairwise_distances(P, Q)  # calculate distance matrix for spectra-peptides
    PxD = process.pairwise_distances(P, D)
    
    # # Get masks
    p_len, q_len = len(P), len(Q)
    PQ_mask, _ = get_masks(counts, p_len, q_len)
    PQ_mask = PQ_mask.to(device)

    # Mine hardest positives
    PxQ_masked = PxQ * PQ_mask
    PxQ_max = Q[PxQ_masked.max(1).indices]

    delta_plus = PxQ_masked.max(1).values #this one

    # # Mine hardest negatives
    PD = torch.cat((P, D))
    PxP.fill_diagonal_(float("inf"))
    PxPD = torch.cat((PxP, PxD), axis=1)
    PxPD_min = PD[PxPD.min(1).indices]              # nearest peptide for each peptide

    delta_minus_P = PxPD.min(1).values # this one

    PQ_neg_mask = (1 - PQ_mask).to(device)
    PxQ_masked = PxQ * PQ_neg_mask
    PxQ_masked[PQ_neg_mask < 0.1] = float("inf")
    PxQ_min = Q[PxQ_masked.min(1).indices]          # nearest spectrum for each peptide
    
    delta_minus_Q = PxQ_masked.min(1).values # this one

    margin = config.get_config(section="ml", key="margin")
    filt1 = ((delta_plus - delta_minus_P)) > 0
    # if len(delta_plus) > torch.sum(filt1):
    #     print("filt1 length: {}".format(torch.sum(filt1).item()))
    A1 = P#[filt1]
    P1 = PxQ_max#[filt1]
    N1 = PxPD_min#[filt1]

    filt2 = ((delta_plus - delta_minus_Q)) > 0
    # if len(delta_plus) > torch.sum(filt2):
    #     print("filt2 length: {}".format(torch.sum(filt2).item()))
    #     print("out of {}".format(len(delta_plus)))
    A2 = P#[filt2]
    P2 = PxQ_max#[filt2]
    N2 = PxQ_min#[filt2]

    # # Calculate the loss
    loss =  triplet_loss(A1, P1, N1)       # peptide-peptide negatives
    loss += triplet_loss(A2, P2, N2)       # peptide-spectrum negatives  
    loss = loss / 2

    return loss, process.pairwise_distances(Q, PD)


# TODO: change it. taken from 
# https://towardsdatascience.com/pytorch-tabular-multiclass-classification-9f8211a123ab
# accessed: 09/18/2020
def multi_acc(y_pred, y_test):
    y_pred_softmax = torch.log_softmax(y_pred, dim = 1)
    _, y_pred_tags = torch.max(y_pred_softmax, dim = 1)    
    
    correct_pred = (y_pred_tags == y_test).float().sum()
    
    return correct_pred


import torch

from specollate.src.snapconfig import config


def generate_percolator_input(l_pep_inds, l_pep_vals, pd_dataset, spec_dataset, res_type):
    assert res_type == "target" or res_type == "decoy"
    assert len(l_pep_inds) == len(l_pep_vals)
    pin_charge = config.get_config(section="search", key="charge")
    l_global_out = []
    tot_count = 0
    max_snap = torch.max(l_pep_vals).item()
    for idx, (pep_inds_row, pep_vals_row) in enumerate(zip(l_pep_inds, l_pep_vals)):
        # Reminder: pep_inds_row length is one less than pep_vals_row
        for iidx in range(len(pep_inds_row) - 1):
            pep_ind = pep_inds_row[iidx]
            pep_val = pep_vals_row[iidx]
            if pep_val.item() > 0:
                charge = [0] * pin_charge
                ch_idx = min(spec_dataset.spec_charge_list[idx], 5)
                spec_idx = spec_dataset.spec_ids[idx]
                charge[ch_idx - 1] = 1
                label = 1 if res_type == "target" else -1
                out_row = [f"{res_type}-{tot_count}", label, spec_idx, pep_val.item()]
                spec_mass = spec_dataset.spec_mass_list[idx]
                pep_mass = pd_dataset.pep_mass_list[pep_ind.item()]
                out_row.append(spec_mass)
                out_row.append(pep_mass)
#                 out_row.append(((pep_val - pep_vals_row[iidx + 1]).item()) / max_snap)
#                 out_row.append(((pep_val - pep_vals_row[-1]).item()) / max_snap)
                out_row.append(((pep_val - pep_vals_row[-3]).item()) / max(pep_val.item(), 1.0))
                out_row.append(((pep_val - pep_vals_row[iidx + 1]).item()) / max(pep_val.item(), 1.0))
                out_row.extend(charge)
                out_row.append(spec_mass - pep_mass)
                out_row.append(abs(spec_mass - pep_mass))
                
                out_pep = pd_dataset.pep_list[pep_ind.item()]
                out_pep_array = []
                for aa in out_pep:
                    if aa.islower():
                        out_pep_array.append("["+str(config.AAMass[aa])+"]")
                    else:
                        out_pep_array.append(aa)
                out_pep = "".join(out_pep_array)
                out_row.append((out_pep.count("K") + out_pep.count("R")) - (out_pep.count("KP") + out_pep.count("RP")) - 1)
                out_prot = pd_dataset.prot_list[pep_ind.item()]
                pep_len = sum([a.isupper() for a in out_pep])
                out_row.append(pep_len)
                out_row.append(out_pep)
                out_row.append(out_prot)
                l_global_out.append(out_row)
                tot_count += 1
    return l_global_out


from os import listdir
from os.path import isfile, join
from pathlib import Path
import shutil
import re

import numpy as np
import torch

from specollate.src.snapconfig import config


def create_out_dir(dir_path, exist_ok=True):
    out_path = Path(dir_path)
    if out_path.exists() and out_path.is_dir():
        if not exist_ok:
            shutil.rmtree(out_path)
            out_path.mkdir()
    else:
        out_path.mkdir()
        
    Path(join(out_path, 'spectra')).mkdir()
    Path(join(out_path, 'peptides')).mkdir()


def verify_in_dir(dir_path, ext, ignore_list=[]):
    in_path = Path(dir_path)
    assert in_path.exists() and in_path.is_dir()
    
    files = [join(dir_path, f) for f in listdir(dir_path) if
                 isfile(join(dir_path, f)) and not f.startswith('.') 
                 and f.split('.')[-1] == ext and f not in ignore_list]
    assert len(files) > 0
    return files


def isfloat(str_float):
    try:
        float(str_float)
        return True
    except ValueError: 
        return False


def mod_repl(match):
    lookup = str(round(float(match.group(0)), 2))
    return config.ModCHAR[lookup] if lookup in config.ModCHAR else ""


def mod_repl_2(match):
    return '[' + str(round(float(match.group(0)), 2)) + ']'


def preprocess_msps(msp_dir, out_dir):
    in_path = Path(msp_dir)
    assert in_path.exists() and in_path.is_dir()
    
    msp_files = [join(msp_dir, f) for f in listdir(msp_dir) if
                 isfile(join(msp_dir, f)) and f.split('.')[-1] == 'msp']
    assert len(msp_files) > 0
    
    out_path = Path(out_dir)
    if out_path.exists() and out_path.is_dir():
        shutil.rmtree(out_path)
    out_path.mkdir()
    Path(join(out_path, 'spectra')).mkdir()
    Path(join(out_path, 'peptides')).mkdir()
        
    print('reading {} files'.format(len(msp_files)))
    
    count = 0
    max_peaks = max_moz = 0
    for species_id, msp_file in enumerate(msp_files):
        print('Reading: {}'.format(msp_file))
        
        f = open(msp_file, "r")
        lines = f.readlines()
        f.close()

        # FIXME: config should use only one get_config call.
        spec_size = config.get_config(section='input', key='spec_size')
        seq_len = config.get_config(section='ml', key='pep_seq_len')

        print('len of file: ' + str(len(lines)))
        limit = 200000
        pep = []
        spec = []
        is_name = is_mw = is_num_peaks = False
        prev = 0
        i = 0
        while i < len(lines) and limit > 0:
            line = lines[i]
            i += 1
            if line.startswith('Name:'):
                name_groups = re.search(r"Name:\s(?P<pep>[a-zA-Z]+)/(?P<charge>\d+)"
                                        r"(?:_(?P<num_mods>\d+)(?P<mods>.*))?", line)
                if not name_groups:
                    continue
                    
                pep = name_groups['pep']
                if len(pep) + 1 > seq_len:
                    continue
                    
                l_charge = int(name_groups['charge'])
                num_mods = int(name_groups['num_mods'])

                is_name = True

            if is_name and line.startswith('MW:'):
                mass = float(re.findall(r"MW:\s([-+]?[0-9]*\.?[0-9]*)", line)[0])
                if round(mass) < spec_size:
                    is_mw = True
                    # limit = limit - 1
                else:
                    is_name = is_mw = is_num_peaks = False
                    continue

            if is_name and is_mw and line.startswith('Num peaks:'):
                num_peaks = int(re.findall(r"Num peaks:\s([0-9]*\.?[0-9]*)", line)[0])
                if num_peaks > max_peaks:
                    max_peaks = num_peaks

                spec = np.zeros(spec_size)
                while lines[i] != '\n':
                    mz_line = lines[i]
                    i += 1
                    mz_splits = mz_line.split('\t')
                    moz, intensity = float(mz_splits[0]), float(mz_splits[1])
                    if moz > max_moz:
                        max_moz = moz
                    spec[round(moz)] += round(intensity)

                # for k in range(1, charge + 1):
                #     spec[-k] = 0
                # spec[-l_charge] = 1000.0
                spec = np.clip(spec, None, 1000.0)
                # spec = preprocessing.scale(spec)

                is_num_peaks = True

            if is_name and is_mw and is_num_peaks:
                is_name = is_mw = is_num_peaks = False
                
                #pep = '{}{}{}'.format(charge, species_id, pep)

                """output the data to """
                spec_tensor = torch.tensor((np.asarray(spec) - 3.725) / 51.479, dtype=torch.float)
                
                torch.save(spec_tensor, 
                           join(out_dir, 'spectra', '{}-{}-{}-{}-{}.pt'
                                .format(count, species_id, mass, l_charge, int(num_mods > 0))))
                
                pep_file_name = '{}-{}-{}-{}-{}.pep'.format(count, species_id, mass, l_charge, int(num_mods > 0))
                    
                with open(join(out_path, 'peptides', pep_file_name), 'w+') as f:
                    f.write(pep)

                count = count + 1
                pep = 0
                spec = []
                new = int((i / len(lines)) * 100)
                if new > prev + 10:
                    # clear_output(wait=True)
                    print(str(new) + '%')
                    prev = new

        print('max peaks: ' + str(max_peaks))
        print('count: ' + str(count))
        print('max moz: ' + str(max_moz))


def preprocess_mgfs(mgf_dir, out_dir):
    
    mgf_files = verify_in_dir(mgf_dir, "mgf")
    create_out_dir(out_dir, exist_ok=False)
        
    print('reading {} files'.format(len(mgf_files)))
    
    spec_size = config.get_config(section='input', key='spec_size')
    charge = config.get_config(section='input', key='charge')
    use_mods = config.get_config(section='input', key='use_mods')
    num_species = config.get_config(section='input', key='num_species')
    seq_len = config.get_config(section='ml', key='pep_seq_len')
    
    ch = np.zeros(20)
    modified = 0
    unmodified = 0
    unique_pep_set = set()
    
    pep_dict = {}
    idx_spec_map = []
    pep_spec = []
    pep_idx = 0
    
    summ = np.zeros(spec_size)
    sq_sum = np.zeros(spec_size)
    N = 0
    
    tot_count = 0
    max_peaks = max_moz = 0
    for species_id, mgf_file in enumerate(mgf_files):
        print('Reading: {}'.format(mgf_file))
        
        f = open(mgf_file, "r")
        lines = f.readlines()
        f.close()
        
        count = lcount = 0
        
        pep_list = []
        dataset = []
        label = []
        
        mass_ign = 0
        pep_len_ign = 0
        dup_ign = 0

        print('len of file: ' + str(len(lines)))
        limit = 200000
        pep = []
        spec = []
        pep_set = set()
        is_name = is_mw = is_charge = is_seq = False
        prev = 0
        i = 0
        while i < len(lines) and limit > 0:
            line = lines[i]
            i += 1

            if line.startswith('PEPMASS'):
                count += 1
                mass = float(re.findall(r"PEPMASS=([-+]?[0-9]*\.?[0-9]*)", line)[0])
                if round(mass)*10 < spec_size:
                    is_mw = True
                    # limit = limit - 1
                else:
                    is_name = is_mw = is_charge = is_seq = False
                    mass_ign += 1
                    continue
            
            if is_mw and line.startswith('CHARGE'):
                l_charge = int(re.findall(r"CHARGE=([-+]?[0-9]*\.?[0-9]*)", line)[0])
                is_charge = True
                mass = (mass - config.PROTON) * l_charge
                
            if is_mw and is_charge and line.startswith("SEQ"):
                line = re.sub(r"[()]", "", line.strip()).split('=')[-1]
                mod_repl_rex = r'([-+]?\d*\.\d+|[-+]?\d+)'
                pep, num_mods = re.subn(mod_repl_rex, mod_repl_2, line)
                is_seq = True
                
            if is_mw and is_charge and is_seq: 
                ind = [] # setting the precision to one decimal point.
                val = []
                for ch_val in range(l_charge):
                    ind.append(ch_val)
                    val.append(1)

                while not isfloat(re.split(' |\t|=', lines[i])[0]):
                    i += 1
                num_peaks = 0   
                while 'END IONS' not in lines[i].upper():
                    if lines[i] == '\n':
                        i += 1
                        continue
                    mz_line = lines[i]
                    i += 1
                    num_peaks += 1
                    mz_splits = re.split(' |\t', mz_line)
                    moz, intensity = float(mz_splits[0]), float(mz_splits[1])
                    if moz > max_moz:
                        max_moz = moz
                    if 0 < round(moz*10) < spec_size:
                        # spec[round(moz*10)] += round(intensity)
                        if ind[-1] == moz*10:
                            val[-1] += intensity
                        else:
                            ind.append(round(moz*10))
                            val.append(intensity)
                if num_peaks < 15:
                    is_name = is_mw = is_charge = is_seq = False
                    continue
                ind = np.array(ind)
                val = np.array(val)
                val = (val - np.amin(val)) / (np.amax(val) - np.amin(val))
                for ch_val in range(l_charge):
                    val[ch_val] = 1
                assert len(ind) == len(val)
                spec = np.array([ind, val])
                
                summ[ind] += val
                sq_sum[ind] += val**2
                N += 1

                is_name = True

            if is_name and is_mw and is_charge and is_seq:
                is_name = is_mw = is_charge = is_seq = False

                """output the data to """
                spec_file_name = '{}-{}-{}.npy'.format(lcount, mass, l_charge)
                np.save(join(out_dir, 'spectra', spec_file_name), spec)
                
                pep_file_name = '{}.pep'.format(lcount)
                with open(join(out_dir, 'peptides', pep_file_name), 'w+') as f:
                    f.write(pep)

                lcount += 1
                tot_count += 1
                
                pep = 0
                spec = []
                new = int((i / len(lines)) * 100)
                if new >= prev + 10:
                    #clear_output(wait=True)
                    print('count: ' + str(lcount))
                    print(str(new) + '%')
                    prev = new

        #print('max peaks: ' + str(max_peaks))
        print('In current file, read {} out of {}'.format(lcount, count))
        print("Ignored: large mass: {}, pep len: {}, dup: {}".format(mass_ign, pep_len_ign, dup_ign))
        print('overall running count: ' + str(tot_count))
        print('max moz: ' + str(max_moz))
#         return pep_list, dataset, label
#         tmp_pep_list, tmp_dataset, tmp_labels = read_msp(msp_file, species_id, decoy)
#         pep_list.extend(tmp_dataset)
#         dataset.extend(tmp_dataset)
#         label.extend(tmp_labels)

    # save the map. this will be used to generate masks for hard positive/negative mining during training.
    # np.save(join(out_dir, "idx_spec_map.npy"), idx_spec_map)
    # with open(join(out_dir, 'pep_spec.pkl'), 'wb') as f:
    #     pickle.dump(pep_spec, f)
    
    print("Statistics:")
    print("Charge distribution:")
    print(ch)
    print("Modified:\t{}".format(modified))
    print("Unmodified:\t{}".format(unmodified))
    print("Unique Peptides:\t{}".format(len(unique_pep_set)))
    print("Sum: {}".format(summ))
    print("Sum-Squared: {}".format(sq_sum))
    print("N: {}".format(N))
    means = summ / N
    print("mean: {}".format(means))
    stds = np.sqrt((sq_sum / N) - means**2)
    stds[stds < 0.0000001] = float("inf")
    print("std: {}".format(stds))
    np.save(join(out_dir, 'means.npy'), means)
    np.save(join(out_dir, 'stds.npy'), stds)

# return spectra, masses, charges


import sys
from os.path import join
from pathlib import Path
import re

import numpy as np
from torch.utils import data
import torch
from tqdm import tqdm

from specollate.src.snapconfig import config
from specollate.src.snaputils import preprocess as prep


class SpectralDataset(data.Dataset):
    'Characterizes a dataset for PyTorch'
    def __init__(self, dir_path):
        'Initialization'
        print(dir_path)
        in_path = Path(dir_path)
        assert in_path.exists()
        assert in_path.is_dir()
        
        self.spec_path = join(dir_path, "spectra")
        self.spec_size = config.get_config(section='input', key='spec_size')
        
        self.means = torch.from_numpy(np.load(join(dir_path, "means.npy"))).float()
        self.stds  = torch.from_numpy(np.load(join(dir_path, "stds.npy"))).float()
        
        spec_ids, spec_lst, spec_mass_lst, spec_charge_lst = load_specs(self.spec_path)
        all_sorts = list(zip(*sorted(zip(spec_ids, spec_lst, spec_mass_lst, spec_charge_lst), key=lambda x: x[2])))
        self.spec_ids         = all_sorts[0]
        self.spec_list        = all_sorts[1]
        self.spec_mass_list   = all_sorts[2]
        self.spec_charge_list = all_sorts[3]
        print('Spectral dataset size: {}'.format(len(self.spec_list)))
        

    def __len__(self):
        'Denotes the total number of samples'
        return len(self.spec_list)


    def __getitem__(self, index):
        'Generates one sample of data'

        # Load spectra
        np_spec = self.spec_list[index]
        ind = torch.LongTensor([[0]*np_spec.shape[1], np_spec[0]])
        val = torch.FloatTensor(np_spec[1])
        torch_spec = torch.sparse_coo_tensor(
            ind, val, torch.Size([1, self.spec_size])).to_dense()
        self.means[:32] = 0.0
        self.stds[:32] = 1.0
        torch_spec = (torch_spec - self.means) / self.stds

        return torch_spec


def load_specs(spec_dir):
    spec_size = config.get_config(key="spec_size", section="input")
    charge = config.get_config(key="charge", section="search")
    spec_files = prep.verify_in_dir(spec_dir, "npy")
    spec_ids = []
    spec_list = []
    masses = []
    charges = []
    count = 0

    pbar = tqdm(spec_files, file=sys.stdout)
    pbar.set_description('Loading Spectra...')
    # with progressbar.ProgressBar(max_value=len(spec_files)) as bar:
    for spec_file in pbar:
        file_name = spec_file.split('/')[-1]
        file_parts = re.search(r"(\d+)-(\d+.\d+)-(\d+).[pt|npy]", file_name)
        spec_id = int(file_parts[1])
        mass = round(float(file_parts[2]), 2)
        l_charge = int(file_parts[3])
        if l_charge > charge:
            continue
        spec_ids.append(spec_id)
        np_spec = np.load(spec_file)
        spec_list.append(np_spec)
        masses.append(mass)
        charges.append(l_charge)

        count += 1
        # bar.update(i)
    # print("count: {}".format(count))
    return spec_ids, spec_list, masses, charges


import json
import os
import tempfile
import shutil
from pathlib import Path
from huggingface_hub import hf_hub_download, snapshot_download
from typing import Optional, Dict, Any

import torch
import pandas as pd

## filter warnings
import warnings
warnings.filterwarnings("ignore")

from .src.snapconfig import config
from .src.snaptrain import model as snap_model
from .src.snapsearch import pepdataset, specdataset, dbsearch, postprocess, preprocess


DEFAULT_CACHE_DIR = Path.home() / ".cache" / "proteorift"


def get_specollate_model_path(cache_dir: Optional[str] = None) -> str:
    """Return path to cached Specollate model or download from HF Hub."""
    cache_dir = cache_dir or DEFAULT_CACHE_DIR
    cache_dir = Path(cache_dir)
    cache_dir.mkdir(parents=True, exist_ok=True)

    # If already cached somewhere under the cache dir, return that
    matches = list(cache_dir.rglob("specollate_model_weights.pt"))
    if matches:
        return str(matches[0])

    repo_default_id = "SaeedLab"
    print("Downloading Specollate model from HuggingFace Hub...")
    specollate_config_path = hf_hub_download(
        repo_id=f"{repo_default_id}/SpeCollate",
        filename="config.json",
        cache_dir=cache_dir,
        repo_type="model"
    )
    ## load file name from config
    with open(specollate_config_path, 'r') as f:
        config = json.load(f)
    model_filename = config.get("model_name", "specollate_model_weights.pt")

    specollate_path = hf_hub_download(
        repo_id=f"{repo_default_id}/SpeCollate",
        filename=model_filename,
        cache_dir=cache_dir,
        repo_type="model"
    )
    print(f"Specollate model downloaded to {specollate_path}")
    return specollate_path


def load_sample_data(cache_dir: Optional[str] = None, use_preprocessed: bool = True) -> Dict[str, str]:
    """Download sample dataset from HuggingFace and return paths.

    Returns a dict with either 'prep_path' and 'peptide_db' (if preprocessed available)
    or 'mgf_dir' and 'peptide_db' for raw files.
    """
    cache_dir = cache_dir or DEFAULT_CACHE_DIR
    cache_dir = Path(cache_dir)
    cache_dir.mkdir(parents=True, exist_ok=True)

    print("Downloading sample data from HuggingFace...")
    repo_path = snapshot_download(repo_id="SaeedLab/sample-data-msms-search", repo_type="dataset", cache_dir=cache_dir)
    repo_path = Path(repo_path)

    if use_preprocessed:
        prep_path = repo_path / "preprocessed"
        peptide_db = repo_path / "raw" / "peptide_database"
        if prep_path.exists() and (prep_path / "specs.pkl").exists():
            print(f"Using preprocessed data: {prep_path}")
            return {"prep_path": str(prep_path), "peptide_db": str(peptide_db)}
        else:
            print("Preprocessed data not found; falling back to raw MGF files")
            use_preprocessed = False

    mgf_dir = repo_path / "raw" / "spectra"
    peptide_db = repo_path / "raw" / "peptide_database"
    return {"mgf_dir": str(mgf_dir), "peptide_db": str(peptide_db)}


class SpeCollateSearch:
    """Simple programmatic interface to run SpeCollate search."""

    def __init__(self, device: str = "auto", precursor_tolerance: float = 7, precursor_tolerance_type: str = "ppm", charge: int = 5):
        if device == "auto":
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device = torch.device(device)
        # store configuration values for use when rendering config files
        self.precursor_tolerance = precursor_tolerance
        self.precursor_tolerance_type = precursor_tolerance_type
        self.charge = charge

    def _create_config_file(self, mgf_dir, prep_dir, pep_dir, out_dir, model_name=None):
        contents = (
            "[search]\n"
            f"mgf_dir = {mgf_dir}\n"
            f"prep_dir = {prep_dir}\n"
            f"pep_dir = {pep_dir}\n"
            f"out_pin_dir = {out_dir}\n"
            f"model_name = {model_name or ''}\n"
            "spec_batch_size = 16384\n"
            "pep_batch_size = 16384\n"
            "search_spec_batch_size = 256\n"
            "keep_psms = 5\n"
            f"precursor_tolerance = {self.precursor_tolerance}\n"
            f"precursor_tolerance_type = {self.precursor_tolerance_type}\n"
            f"charge = {self.charge}\n"
            "\n"
            "[input]\n"
            "spec_size = 80000\n"
            "charge = 5\n"
            "use_mods = False\n"
            "\n"
            "[ml]\n"
            "batch_size = 1024\n"
            "pep_seq_len = 36\n"
            "dropout = 0.3\n"
        )
        tf = tempfile.NamedTemporaryFile(mode="w", suffix=".ini", delete=False)
        tf.write(contents)
        tf.close()
        return tf.name

    def _preprocess_spectra(self, mgf_dir, prep_dir):
        # Create a temporary config file (so snapconfig can read `input` section)
        cfg = self._create_config_file(mgf_dir, prep_dir, "", "", model_name=None)
        # Point the snapconfig module at the new config
        config.PARAM_PATH = cfg
        try:
            # preprocess.mgfs expects (mgf_dir, prep_dir)
            preprocess.preprocess_mgfs(mgf_dir, prep_dir)
        finally:
            try:
                os.unlink(cfg)
            except Exception:
                pass

    def search(
        self,
        mgf_dir: str,
        peptide_db: str,
        output_dir: str = "./specollate_output",
        prep_dir: Optional[str] = None,
        use_cached_prep: bool = False,
        model_path: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Run preprocessing (if needed) and database search.

        Returns paths to generated `target.pin` and `decoy.pin`.
        """
        os.makedirs(output_dir, exist_ok=True)

        cleanup_prep = False
        if prep_dir is None:
            prep_dir = tempfile.mkdtemp(prefix="specollate_prep_")
            cleanup_prep = True
        else:
            os.makedirs(prep_dir, exist_ok=True)

        prep_file = Path(prep_dir) / "specs.pkl"
        if not use_cached_prep or not prep_file.exists():
            self._preprocess_spectra(mgf_dir, prep_dir)

        # Create a minimal config and point the snapconfig module at it
        # If no model path provided, attempt to download/get cached Specollate weights
        if model_path is None:
            try:
                model_path = get_specollate_model_path()
            except Exception as e:
                print("Warning: failed to download Specollate model:", e)
                model_path = None

        cfg_file = self._create_config_file(mgf_dir, prep_dir, peptide_db, output_dir, model_path)
        config.PARAM_PATH = cfg_file

        # Build datasets and loaders
        spec_ds = specdataset.SpectralDataset(prep_dir)
        pep_ds = pepdataset.PeptideDataset(peptide_db)
        dec_ds = pepdataset.PeptideDataset(peptide_db, decoy=True)

        spec_batch_size = config.get_config(key="spec_batch_size", section="search")
        pep_batch_size = config.get_config(key="pep_batch_size", section="search")
        search_spec_batch_size = config.get_config(key="search_spec_batch_size", section="search")

        spec_loader = torch.utils.data.DataLoader(
            dataset=spec_ds, batch_size=spec_batch_size, collate_fn=dbsearch.spec_collate
        )

        pep_loader = torch.utils.data.DataLoader(
            dataset=pep_ds, batch_size=pep_batch_size, collate_fn=dbsearch.pep_collate
        )

        dec_loader = torch.utils.data.DataLoader(
            dataset=dec_ds, batch_size=pep_batch_size, collate_fn=dbsearch.pep_collate
        )

        # Load model
        vocab_size = 30
        embedding_dim = 512
        hidden_lstm_dim = 512
        lstm_layers = 2
        net = snap_model.Net(vocab_size=vocab_size, embedding_dim=embedding_dim,
                             hidden_lstm_dim=hidden_lstm_dim, lstm_layers=lstm_layers).to(self.device)
        model_name = config.get_config(key="model_name", section="search")
        if model_path:
            model_name = model_path
        if model_name:
            state = torch.load(model_name, map_location=self.device)
            if isinstance(state, dict) and 'model_state_dict' in state:
                sd = state['model_state_dict']
            elif isinstance(state, dict):
                sd = state
            else:
                raise RuntimeError("Loaded model file does not contain a state dict")

            cleaned = {}
            for k, v in sd.items():
                if k.startswith('module.'):
                    cleaned[k[len('module.'):]] = v
                else:
                    cleaned[k] = v

            net.load_state_dict(cleaned)

        # runModel expects model in eval mode
        net.eval()

        # Run embeddings
        e_specs = dbsearch.runModel(spec_loader, net, "specs", self.device)
        e_peps = dbsearch.runModel(pep_loader, net, "peps", self.device)
        e_decs = dbsearch.runModel(dec_loader, net, "peps", self.device)

        # Run search
        search_loader = torch.utils.data.DataLoader(dataset=e_specs, batch_size=search_spec_batch_size, shuffle=False)
        datasets = {"spec_dataset": spec_ds, "pep_dataset": pep_ds, "dec_dataset": dec_ds}
        embeddings = {"e_specs": e_specs, "e_peps": e_peps, "e_decs": e_decs}

        pep_inds, pep_vals, dec_inds, dec_vals = dbsearch.search(search_loader, datasets, embeddings, self.device)

        # Write percolator files
        pin_charge = config.get_config(section="search", key="charge")
        charge_cols = [f"charge-{ch+1}" for ch in range(pin_charge)]
        cols = ["SpecId", "Label", "ScanNr", "SNAP", "ExpMass", "CalcMass", "deltCn", "deltLCn"] + charge_cols + ["dM", "absdM", "enzInt", "PepLen", "Peptide", "Proteins"]

        global_out = postprocess.generate_percolator_input(pep_inds, pep_vals, pep_ds, spec_ds, "target")
        df = pd.DataFrame(global_out, columns=cols)
        df.sort_values(by="SNAP", inplace=True, ascending=False)
        df.to_csv(os.path.join(output_dir, "target.pin"), sep="\t", index=False)

        global_out = postprocess.generate_percolator_input(dec_inds, dec_vals, dec_ds, spec_ds, "decoy")
        df = pd.DataFrame(global_out, columns=cols)
        df.sort_values(by="SNAP", inplace=True, ascending=False)
        df.to_csv(os.path.join(output_dir, "decoy.pin"), sep="\t", index=False)

        # cleanup
        try:
            os.unlink(cfg_file)
        except Exception:
            pass
        if cleanup_prep:
            shutil.rmtree(prep_dir)
        
        print(f"Search completed. Outputs written to {output_dir}")

        return {
            "output_dir": output_dir,
            "target_pin": os.path.join(output_dir, "target.pin"),
            "decoy_pin": os.path.join(output_dir, "decoy.pin"),
        }

    def search_with_sample_data(self, output_dir: str = "./specollate_output", use_preprocessed: bool = True, cache_dir: Optional[str] = None) -> Dict[str, Any]:
        """Run search using the sample dataset hosted on HuggingFace.

        This will download the sample dataset (preprocessed if available) and run
        the same `search` pipeline, reusing preprocessed data when present.
        """
        print("Loading sample data from HuggingFace...")
        sample_data = load_sample_data(cache_dir=cache_dir, use_preprocessed=use_preprocessed)

        # If preprocessed data available, use it directly by pointing `prep_dir`
        if 'prep_path' in sample_data:
            prep_path = sample_data['prep_path']
            peptide_db = sample_data['peptide_db']
            # Ensure output dir exists
            os.makedirs(output_dir, exist_ok=True)
            # Use existing search method but skip preprocessing
            self.search(mgf_dir="", peptide_db=peptide_db, output_dir=output_dir, prep_dir=prep_path, use_cached_prep=True)
        else:
            # Raw MGF files: run full search (will preprocess)
            self.search(mgf_dir=sample_data['mgf_dir'], peptide_db=sample_data['peptide_db'], output_dir=output_dir)



from os.path import join
from pathlib import Path
import re

import numpy as np
import torch
from torch.utils import data

from specollate.src.snapconfig import config


class LabeledSpectra(data.Dataset):
    'Characterizes a dataset for PyTorch'
    def __init__(self, dir_path, pep_file_name, spec_file_names_lists):
        'Initialization'
        
        in_path = Path(dir_path)
        assert in_path.exists()
        assert in_path.is_dir()
        self.aas            = ['_PAD'] + list(config.AAMass.keys())# + list(config.ModCHAR.values())
        self.aa2idx         = {a:i for i, a in enumerate(self.aas)}
        self.idx2aa         = {i:a for i, a in enumerate(self.aas)}
        self.charge         = config.get_config(section='input', key='charge')
        self.charge2idx     = {i+1:i for i in range(self.charge)}
        self.idx2charge     = {i:i+1 for i in range(self.charge)}
        
        self.spec_path      = join(dir_path, 'spectra')
        self.pep_path       = join(dir_path, 'peptides')
        self.num_species    = config.get_config(section='input', key='num_species')
        self.vocab_size     = len(self.aa2idx) # + self.charge + self.num_species + 1
        print("Vocabulary Size: {}".format(self.vocab_size))
        # self.vocab_size   = round(max(config.AAMass.values())) + 1
        self.spec_size      = config.get_config(section='input', key='spec_size')
        self.seq_len        = config.get_config(section='ml', key='pep_seq_len')
        
        self.pep_file_names = pep_file_name
        self.spec_file_names_lists = spec_file_names_lists # a list of lists containing spectra for each peptide
        
        means = np.load(join(dir_path, "means.npy"))
        stds = np.load(join(dir_path, "stds.npy"))
        self.means = torch.from_numpy(means).float()
        self.stds = torch.from_numpy(stds).float()

        # aux_means = np.load(join(dir_path, "aux_means.npy"))
        # aux_stds = np.load(join(dir_path, "aux_stds.npy"))
        # self.aux_means = torch.from_numpy(aux_means).float()
        # self.aux_stds = torch.from_numpy(aux_stds).float()

        # db_peps_path = config.get_config(section="input", key="db_peps_path")
        # with open(db_peps_path, 'rb') as fp:
        #     self.db_peps = pickle.load(fp)
        
        print('dataset size: {}'.format(len(self.pep_file_names)))
        

    def __len__(self):
        'Denotes the total number of samples'
        return len(self.pep_file_names)


    def __getitem__(self, index):
        'Generates one sample of data'
        pep_file_name = ''
        spec_file_list = []
        # Select sample
        pep_file_name = self.pep_file_names[index]
        spec_file_list = self.spec_file_names_lists[index]

        # Load spectra
        torch_spec_list = []
        torch_spec_charge_list = []
        torch_spec_mass_list = []
        for spec_file in spec_file_list:
            file_parts = re.search(r"(\d+)-(\d+)-(\d+.\d+)-(\d+)-(\d+).[pt|npy]", spec_file)
            spec_charge = int(file_parts[4])
            spec_mass = float(file_parts[3])
            np_spec = np.load(join(self.spec_path, spec_file))
            ind = torch.LongTensor([[0]*np_spec.shape[1], np_spec[0]])
            val = torch.FloatTensor(np_spec[1])

            # coin_toss = random.randrange(0,2,1)
            # if coin_toss == 1:
            #     zero_inds = random.sample(range(len(val)), int(len(val)*0.1))
            #     val[zero_inds] = 0.0
            
            torch_spec = torch.sparse_coo_tensor(
                ind, val, torch.Size([1, self.spec_size])).to_dense()
            # self.means[:32] = 0.0
            # self.stds[:32] = 1.0
            torch_spec = (torch_spec - self.means) / self.stds
            # torch_spec[1:10] = 0.
            torch_spec_list.append(torch_spec)
            torch_spec_charge_list.append(spec_charge)
            torch_spec_mass_list.append(spec_mass)
        
        torch_spec_charge_list = [self.charge2idx[charge] for charge in torch_spec_charge_list]
        torch_spec = torch.cat(torch_spec_list, dim=0)
        torch_spec_charge = torch.LongTensor(torch_spec_charge_list)
        torch_spec_mass = torch.FloatTensor(torch_spec_mass_list)

        # Load peptide
        pep_file_name = join(self.pep_path, pep_file_name)
        f = open(pep_file_name, "r")
        pep = f.readlines()[0].strip()
        f.close()
        
        pepl = [self.aa2idx[aa] for aa in pep]
        # torch_pep = self.one_hot_tensor(pepl) # no need to call pad after this.
        pepl = self.pad_left(pepl, self.seq_len)

        # pep_mass = sim.get_pep_mass(pep)
        # gray_mass = sim.gray_code(round(pep_mass * 1000))
        # mass_arr = sim.decimal_to_binary_array(gray_mass, 27)
        # pepl = np.concatenate((mass_arr, pepl))

        # pep_len = np.zeros(64)
        # missed_cleavs_vec = np.zeros(17)
        # aas = np.zeros(30)
        # pep_len[len(pep)] = 1
        # missed_cleavs = (pep.count("K") + pep.count("R")) - (pep.count("KP") + pep.count("RP")) - 1 # might not be correct for non-tryp peps
        # missed_cleavs_vec[missed_cleavs] = 1
        # for a in pep:
        #     aa_id = self.aa2idx[a]
        #     aas[aa_id] += 1
        # concat_vec = np.concatenate((pep_len, missed_cleavs_vec, aas))
        # concat_vec = (concat_vec - self.aux_means) / self.aux_stds
        # pepl = np.concatenate((concat_vec, pepl))
        
        torch_pep = torch.tensor(pepl, dtype=torch.long)
        dpep = self.get_decoy(pep)
        # dpep = random.choice(self.db_peps[round(pep_mass*10)]) if self.db_peps[round(pep_mass*10)] else self.get_decoy(pep)
        if dpep:
            dpepl = [self.aa2idx[aa] for aa in dpep]
            # torch_pep_decoy = self.one_hot_tensor(dpepl) # no need to call pad after this.
            dpepl = self.pad_left(dpepl, self.seq_len)
            # dpepl = np.concatenate((mass_arr, dpepl))
            # dpepl = np.concatenate((concat_vec, dpepl))
            torch_pep_decoy = torch.tensor(dpepl, dtype=torch.long)
        else:
            torch_pep_decoy = torch.empty(0, 0, dtype=torch.long)
        # return (torch_spec, torch_pep, torch_pep_decoy, torch_spec_charge,
        #     torch_spec_mass, pep_mass, len(torch_spec_list))
        return (torch_spec, torch_pep, torch_pep_decoy, torch_spec_charge, len(torch_spec))

    def pad_left(self, arr, size):
        out = np.zeros(size)
        out[-len(arr):] = arr
        return out


    # TODO: Needs fixing for multiple n-term mods
    def get_decoy(self, pep):
        first = ""
        if pep[0].islower():
            first = pep[0]
            pep = pep[1:]
        
        pep_parts = re.findall(r"([A-Z][a-z]?)", pep)
        decoy_pep = pep_parts[0] + "".join(pep_parts[-2:0:-1]) + pep_parts[-1] # peptide reverse
        # middle = pep_parts[1:-1]
        # random.shuffle(middle)
        # decoy_pep = pep_parts[0] + "".join(middle) + pep_parts[-1] # peptide shuffle
        decoy_pep = first + decoy_pep
        if decoy_pep == pep:
            decoy_pep = []

        return decoy_pep


    # def one_hot_tensor(self, pep):
    #     one_hots = torch.zeros((len(pep), self.vocab_size), dtype=torch.long)
    #     zeros = torch.zeros(((self.seq_len - len(pep)), self.vocab_size), dtype=torch.long)
    #     index = torch.tensor(pep)
    #     one_hots.scatter_(1, index.view(-1, 1), 1)
    #     return torch.cat((zeros, one_hots), dim=0)




from os import listdir
from os.path import isfile, join
from pathlib import Path
import shutil
import re
import math

import numpy as np

from specollate.src.snapconfig import config


def create_out_dir(dir_path, exist_ok=True):
    out_path = Path(dir_path)
    if out_path.exists() and out_path.is_dir():
        if not exist_ok:
            shutil.rmtree(out_path)
            out_path.mkdir()
    else:
        out_path.mkdir()
        
    Path(join(out_path, 'spectra')).mkdir()
    Path(join(out_path, 'peptides')).mkdir()


def verify_in_dir(dir_path, ext, ignore_list=[]):
    in_path = Path(dir_path)
    assert in_path.exists() and in_path.is_dir()
    
    files = [join(dir_path, f) for f in listdir(dir_path) if
                 isfile(join(dir_path, f)) and not f.startswith('.') 
                 and f.split('.')[-1] == ext and f not in ignore_list]
    assert len(files) > 0
    return files


def isfloat(str_float):
    try:
        float(str_float)
        return True
    except ValueError: 
        return False


def mod_repl(match):
    lookup = str(round(float(match.group(0)), 2))
    return config.ModCHAR[lookup] if lookup in config.ModCHAR else ""


def mod_repl_2(match):
    return '[' + str(round(float(match.group(0)), 2)) + ']'


def preprocess_mgfs(mgf_dir, out_dir):
   
    mgf_files = verify_in_dir(mgf_dir, "mgf")
    create_out_dir(out_dir, exist_ok=False)
        
    print('reading {} files'.format(len(mgf_files)))
    
    spec_size = config.get_config(section='input', key='spec_size')
    
    ch = np.zeros(20)
    modified = 0
    unmodified = 0
    unique_pep_set = set()
    
    summ = np.zeros(spec_size)
    sq_sum = np.zeros(spec_size)
    N = 0
    
    tot_count = 0
    max_moz = 0
    for mgf_file in mgf_files:
        print('Reading: {}'.format(mgf_file))
        
        f = open(mgf_file, "r")
        lines = f.readlines()
        f.close()
        
        count = lcount = 0
        
        mass_ign = 0
        pep_len_ign = 0
        dup_ign = 0

        print('len of file: ' + str(len(lines)))
        limit = 200000
        spec = []
        is_name = is_mw = is_charge = False
        prev = 0
        i = 0
        while i < len(lines) and limit > 0:
            line = lines[i]
            i += 1

            if line.startswith('PEPMASS'):
                count += 1
                mass = float(re.findall(r"PEPMASS=([-+]?[0-9]*\.?[0-9]*)", line)[0])
                is_mw = True
            
            if is_mw and line.startswith('CHARGE'):
                l_charge = int(re.findall(r"CHARGE=([-+]?[0-9]*\.?[0-9]*)", line)[0])
                is_charge = True
                mass = (mass - config.PROTON) * l_charge
                
            if is_mw and is_charge:

                while not isfloat(re.split(' |\t|=', lines[i])[0]):
                    i += 1
                    
                spec_ind = []
                spec_val = []
                num_peaks = 0
                while 'END IONS' not in lines[i].upper():
                    if lines[i] == '\n':
                        i += 1
                        continue
                    mz_line = lines[i]
                    i += 1
                    num_peaks += 1
                    mz_splits = re.split(' |\t', mz_line)
                    moz = round(float(mz_splits[0]) * 10) # + 32 # 32 because charge is len 8 and mass is len 24
                    intensity = math.sqrt(float(mz_splits[1]) + 1.0) # adding 1 to avoid sqrt of zero
#                     intensity = float(mz_splits[1])
                    if moz > max_moz:
                        max_moz = moz
                    if 0 < moz < spec_size:
                        # spec[round(moz*10)] += round(intensity)
                        if spec_ind and spec_ind[-1] == moz:
                            spec_val[-1] = max(intensity, spec_val[-1])
                        else:
                            spec_ind.append(moz)
                            spec_val.append(intensity) # adding one to avoid sqrt of zero
                if num_peaks < 10:
                    is_name = is_mw = is_charge = False
                    continue
                    
                spec_ind = np.array(spec_ind)
                spec_val = np.array(spec_val)
                spec_val = (spec_val - np.amin(spec_val)) / (np.amax(spec_val) - np.amin(spec_val))

                ind = spec_ind
                val = spec_val
            
                assert len(ind) == len(val)
                spec = np.array([ind, val])
                
                summ[ind] += val
                sq_sum[ind] += val**2
                N += 1

                is_name = True

            if is_name and is_mw and is_charge:
                is_name = is_mw = is_charge = False

                """output the data to """
                spec_file_name = '{}-{}-{}.npy'.format(lcount, mass, l_charge)
                np.save(join(out_dir, 'spectra', spec_file_name), spec)

                lcount += 1
                tot_count += 1
                
                pep = 0
                spec = []
                new = int((i / len(lines)) * 100)
                if new >= prev + 10:
                    #clear_output(wait=True)
                    print('count: ' + str(lcount))
                    print(str(new) + '%')
                    prev = new

        #print('max peaks: ' + str(max_peaks))
        print('In current file, read {} out of {}'.format(lcount, count))
        print("Ignored: large mass: {}, pep len: {}, dup: {}".format(mass_ign, pep_len_ign, dup_ign))
        print('overall running count: ' + str(tot_count))
        print('max m/z: ' + str(max_moz))
    
    # print("Statistics:")
    # print("Charge distribution:")
    # print(ch)
    # print("Modified:\t{}".format(modified))
    # print("Unmodified:\t{}".format(unmodified))
    # print("Unique Peptides:\t{}".format(len(unique_pep_set)))
    # print("Sum: {}".format(summ))
    # print("Sum-Squared: {}".format(sq_sum))
    # print("N: {}".format(N))
    means = summ / N
    # print("mean: {}".format(means))
    stds = np.sqrt((sq_sum / N) - means**2)
    stds[stds < 0.0000001] = float("inf")
    # print("std: {}".format(stds))
    np.save(join(out_dir, 'means.npy'), means)
    np.save(join(out_dir, 'stds.npy'), stds)

# return spectra, masses, charges


import argparse
import pandas as pd
import os
from tqdm import tqdm

import torch
import torch.nn as nn
import torch.distributed as dist
from os.path import join

from .src.snapconfig import config
from .src.snaptrain import model
from .src.snapsearch import pepdataset, specdataset, dbsearch, postprocess, preprocess


if __name__ == '__main__':
    # Initialize parser 
    parser = argparse.ArgumentParser()
    
    # Adding optional argument 
    parser.add_argument("-c", "--config", help="Path to the config file.")
    parser.add_argument("-p", "--preprocess", help="Which server the code is running on. \
        Options: raptor, comet. Default: comet", default="True")
    
    # Read arguments from command line 
    args = parser.parse_args() 

    if args.config:
        tqdm.write("config: %s" % args.path)
    # config.PARAM_PATH = args.config if args.config else join((os.path.dirname(__file__)), "config.ini")
    
    mgf_dir     = config.get_config(key="mgf_dir", section="search")
    prep_dir    = config.get_config(key="prep_dir", section="search")
    pep_dir     = config.get_config(key="pep_dir", section="search")
    out_pin_dir = config.get_config(key="out_pin_dir", section="search")

    if not os.path.exists(out_pin_dir):
        os.mkdir(out_pin_dir)

    # scratch_loc = "/scratch/mtari008/job_" + os.environ['SLURM_JOB_ID'] + "/"

    # mgf_dir     = scratch_loc + mgf_dir
    # prep_dir    = scratch_loc + prep_dir
    # pep_dir     = scratch_loc + pep_dir
    # out_pin_dir = scratch_loc + out_pin_dir
    
    if args.preprocess and args.preprocess == "True":
        tqdm.write("Preprocessing mgf files...")
        preprocess.preprocess_mgfs(mgf_dir, prep_dir)

    tqdm.write("Reading input files...")
    spec_dataset = specdataset.SpectralDataset(prep_dir)
    pep_dataset  = pepdataset.PeptideDataset(pep_dir)
    dec_dataset  = pepdataset.PeptideDataset(pep_dir, decoy=True)

    spec_batch_size = config.get_config(key="spec_batch_size", section="search")
    pep_batch_size  = config.get_config(key="pep_batch_size", section="search")

    print("Generating data loaders...")
    spec_loader = torch.utils.data.DataLoader(
        dataset=spec_dataset, batch_size=spec_batch_size,
        collate_fn=dbsearch.spec_collate)
    pep_loader = torch.utils.data.DataLoader(
        dataset=pep_dataset, batch_size=pep_batch_size,
        collate_fn=dbsearch.pep_collate)
    dec_loader = torch.utils.data.DataLoader(
        dataset=dec_dataset, batch_size=pep_batch_size,
        collate_fn=dbsearch.pep_collate)

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    print(device)
    os.environ['MASTER_ADDR'] = 'localhost'
    os.environ['MASTER_PORT'] = '12350'
    dist.init_process_group(backend='nccl', world_size=1, rank=0)
    snap_model = model.Net(vocab_size=30, embedding_dim=512, hidden_lstm_dim=512, lstm_layers=2).to(device)
    # print(snap_model.linear1_1.weight.shape)
    snap_model = nn.parallel.DistributedDataParallel(snap_model, device_ids=[0])
    # snap_model.load_state_dict(torch.load('models/32-embed-2-lstm-SnapLoss2-noch-3k-1k-152.pt')['model_state_dict'])
    # below one has 26975 identified peptides.
    # snap_model.load_state_dict(torch.load('models/512-embed-2-lstm-SnapLoss-noch-80k-nist-massive-52.pt')['model_state_dict'])
    model_name =  config.get_config(key="model_name", section="search")
    snap_model.load_state_dict(torch.load(model_name)['model_state_dict'])
    snap_model = snap_model.module
    snap_model.eval()
    snap_model

    print("Processing spectra...")
    e_specs = dbsearch.runModel(spec_loader, snap_model, "specs", device)
    print("Spectra done!")

    print("Processing peptides...")
    e_peps = dbsearch.runModel(pep_loader, snap_model, "peps", device)
    print("Peptides done!")

    print("Processing decoys...")
    e_decs = dbsearch.runModel(dec_loader, snap_model, "peps", device)
    print("Decoys done!")

    search_spec_batch_size = config.get_config(key="search_spec_batch_size", section="search")
    datasets = {"spec_dataset":spec_dataset, "pep_dataset":pep_dataset, "dec_dataset":dec_dataset}
    embeddings = {"e_specs":e_specs, "e_peps":e_peps, "e_decs":e_decs}
    search_loader = torch.utils.data.DataLoader(
        dataset=e_specs, batch_size=search_spec_batch_size, shuffle=False)
    
    pep_inds, pep_vals, dec_inds, dec_vals = dbsearch.search(search_loader, datasets, embeddings, device)

    pin_charge = config.get_config(section="search", key="charge")
    charge_cols = [f"charge-{ch+1}" for ch in range(pin_charge)]
    cols = ["SpecId", "Label", "ScanNr", "SNAP", "ExpMass", "CalcMass", "deltCn", "deltLCn"] + charge_cols + ["dM", "absdM", "enzInt", "PepLen", "Peptide", "Proteins"]
    
    print("Generating percolator pin files...")
    global_out = postprocess.generate_percolator_input(pep_inds, pep_vals, pep_dataset, spec_dataset, "target")
    df = pd.DataFrame(global_out, columns=cols)
    df.sort_values(by="SNAP", inplace=True, ascending=False)
    df.to_csv(join(out_pin_dir, "target.pin"), sep="\t", index=False)

    global_out = postprocess.generate_percolator_input(dec_inds, dec_vals, dec_dataset, spec_dataset, "decoy")
    df = pd.DataFrame(global_out, columns=cols)
    df.sort_values(by="SNAP", inplace=True, ascending=False)
    df.to_csv(join(out_pin_dir, "decoy.pin"), sep="\t", index=False)
    print("Wrote percolator files: \n{}\n{}".format(
        join(out_pin_dir, "target.pin"), join(out_pin_dir, "decoy.pin")))


import re
from os import listdir
from os.path import isfile, join

import numpy as np
from IPython.display import clear_output
from sklearn import preprocessing

from specollate.src.snapconfig import config
from specollate.src.snaputils import simulatespectra as sim


def read_msps(msp_folder, decoy=False):
    msp_files = [join(msp_folder, f) for f in listdir(msp_folder) if
                 isfile(join(msp_folder, f)) and f.split('.')[-1] == 'msp']
    assert len(msp_files) > 0

    print('reading {} files'.format(len(msp_files)))
    pep_list = []
    dataset = []
    label = []

    for species_id, msp_file in enumerate(msp_files):
        print('Reading: {}'.format(msp_file))
        tmp_pep_list, tmp_dataset, tmp_labels = read_msp(msp_file, species_id, decoy)
        pep_list.extend(tmp_dataset)
        dataset.extend(tmp_dataset)
        label.extend(tmp_labels)

    return pep_list, dataset, label


def read_msp(msp_file, species_id, decoy=False):
    """Read annotated spectra from msp file and return
    peptide list, dataset, and labels.
    :param decoy:
    :param species_id: id of the species
    :param msp_file: str
    :returns list
    """

    f = open(msp_file, "r")
    lines = f.readlines()
    f.close()

    pep_list = []
    dataset = []
    label = []

    # FIXME: config should use only one get_config call.
    spec_size = config.get_config(section='input', key='spec_size')
    charge = config.get_config(section='input', key='charge')
    use_mods = config.get_config(section='input', key='use_mods')
    num_species = config.get_config(section='input', key='num_species')

    print('len of file: ' + str(len(lines)))
    count = 0
    limit = 200000
    pep = []
    spec = []
    pep_set = set()
    is_name = is_mw = is_num_peaks = False
    prev = 0
    max_peaks = max_moz = 0
    i = 0
    while i < len(lines) and limit > 0:
        line = lines[i]
        i += 1
        if line.startswith('Name:'):
            name_groups = re.search(r"Name:\s(?P<pep>[a-zA-Z]+)/(?P<charge>\d+)"
                                    r"(?:_(?P<num_mods>\d+)(?P<mods>.*))?", line)
            if not name_groups:
                continue
            pep = name_groups['pep']
            l_charge = int(name_groups['charge'])
            num_mods = int(name_groups['num_mods'])

            if l_charge > charge:
                continue

            if (use_mods or not num_mods) and pep + str(l_charge) not in pep_set:
                pep_set.add(pep + str(l_charge))
                is_name = True
            else:
                continue

        if is_name and line.startswith('MW:'):
            mass = float(re.findall(r"MW:\s([-+]?[0-9]*\.?[0-9]*)", line)[0])
            if round(mass) < spec_size:
                is_mw = True
                # limit = limit - 1
            else:
                is_name = is_mw = is_num_peaks = False
                continue

        if is_name and is_mw and line.startswith('Num peaks:'):
            num_peaks = int(re.findall(r"Num peaks:\s([0-9]*\.?[0-9]*)", line)[0])
            if num_peaks > max_peaks:
                max_peaks = num_peaks

            spec = np.zeros(spec_size)
            while lines[i] != '\n':
                mz_line = lines[i]
                i += 1
                mz_splits = mz_line.split('\t')
                moz, intensity = float(mz_splits[0]), float(mz_splits[1])
                if moz > max_moz:
                    max_moz = moz
                spec[round(moz)] += round(intensity)

            # for k in range(1, charge + 1):
            #     spec[-k] = 0
            # spec[-l_charge] = 1000.0
            spec = np.clip(spec, None, 1000.0)
            # spec = preprocessing.scale(spec)

            is_num_peaks = True

        if is_name and is_mw and is_num_peaks:
            is_name = is_mw = is_num_peaks = False
            # revPep = pep[0] + pep[1:-1][::-1] + pep[-1]
            pep_list.append(pep)
            t_spec = sim.get_spectrum(pep)

            for k in range(0, charge):
                t_spec[k] = 1.0 if k <= l_charge - 1 else 0.0
            for k in range(charge, charge + num_species):
                t_spec[k] = 1.0 if k - charge == species_id else 0.0
            t_spec = preprocessing.scale(t_spec)

            if decoy:
                revPep = sim.get_rand_mod(pep)
                if pep == revPep:
                    print('decoy is the same. shuffling')
                    # revPep = ''.join(rand.sample(revPep,len(revPep)))
                    revPep = sim.get_rand_mod(pep, len(pep))
                    print(pep)
                    print(revPep)
                rt_spec = sim.get_spectrum(revPep)
                #rt_spec = preprocessing.scale(rt_spec)
                dataset.append([spec, t_spec, rt_spec])
                label.append([1, -1])
            else:
                dataset.append([spec, t_spec])
                label.append([1])

            count = count + 1
            pep = 0
            spec = []
            new = int((i / len(lines)) * 100)
            if new > prev:
                clear_output(wait=True)
                print(str(new) + '%')
                prev = new

    print('max peaks: ' + str(max_peaks))
    print('count: ' + str(count))
    print('max moz: ' + str(max_moz))
    return pep_list, dataset, label


# def read_msp_backup(msp_file, decoy=False):
#     """Read annotated spectra from msp file and return
#     peptide list, dataset, and labels.
#     :param msp_file: str
#     :returns list
#     """
#
#     f = open(msp_file, "r")
#     lines = f.readlines()
#     f.close()
#
#     pep_list = []
#     dataset = []
#     label = []
#
#     # FIXME: config should use only one get_config call.
#     spec_size = config.get_config(section='input', key='spec_size')
#     charge = config.get_config(section='input', key='charge')
#     use_mods = config.get_config(section='input', key='use_mods')
#
#     print('len of file: ' + str(len(lines)))
#     count = 0
#     limit = 200000
#     pep = []
#     spec = []
#     pep_set = set()
#     is_name = is_mw = is_num_peaks = False
#     prev = 0
#     max_peaks = max_moz = 0
#     i = 0
#     while i < len(lines) and limit > 0:
#         line = lines[i]
#         i += 1
#         splits = line.split(':')
#         if (splits[0] == 'Name') and '_' in line:
#             split1 = splits[1]
#             l_charge = int(split1[split1.find('_') - 1])
#             if l_charge != charge and charge > 0:  # l_charge == l_charge always true.
#                 continue
#
#             if use_mods or ('(' not in splits[1] and ')' not in splits[1]):
#                 pep = split1.split('/')[0].lstrip(' ')
#
#                 if pep not in pep_set:
#                     pep_set.add(pep)
#                 else:
#                     continue
#
#                 is_name = True
#
#         if is_name and splits[0] == 'MW':
#             mass = float(splits[1])
#             if round(mass) < spec_size:
#                 is_mw = True
#                 # limit = limit - 1
#             else:
#                 is_name = is_mw = is_num_peaks = False
#                 continue
#
#         if is_name and is_mw and splits[0] == 'Num peaks':
#             num_peaks = int(splits[1])
#             if num_peaks > max_peaks:
#                 max_peaks = num_peaks
#
#             spec = np.zeros(spec_size)
#             while lines[i] != '\n':
#                 mz_line = lines[i]
#                 i += 1
#                 mz_splits = mz_line.split('\t')
#                 moz, intensity = float(mz_splits[0]), float(mz_splits[1])
#                 if moz > max_moz:
#                     max_moz = moz
#                 spec[round(moz)] += round(intensity)
#
#             spec = np.clip(spec, None, 1000.0)
#             spec = preprocessing.scale(spec)
#
#             is_num_peaks = True
#
#         if is_name and is_mw and is_num_peaks:
#             is_name = is_mw = is_num_peaks = False
#             # revPep = pep[0] + pep[1:-1][::-1] + pep[-1]
#             pep_list.append(pep)
#             t_spec = preprocessing.scale(sim.get_spectrum(pep))
#             if decoy:
#                 revPep = sim.get_rand_mod(pep)
#                 if pep == revPep:
#                     print('decoy is the same. shuffling')
#                     # revPep = ''.join(rand.sample(revPep,len(revPep)))
#                     revPep = sim.get_rand_mod(pep, len(pep))
#                     print(pep)
#                     print(revPep)
#                 rt_spec = preprocessing.scale(sim.get_spectrum(revPep))
#                 dataset.append([spec, t_spec, rt_spec])
#                 label.append([1, -1])
#             else:
#                 dataset.append([spec, t_spec])
#                 label.append([1])
#
#             count = count + 1
#             pep = 0
#             spec = []
#             new = int((i / len(lines)) * 100)
#             if new > prev:
#                 clear_output(wait=True)
#                 print(str(new) + '%')
#                 prev = new
#
#     print('max peaks: ' + str(max_peaks))
#     print('count: ' + str(count))
#     print('max moz: ' + str(max_moz))
#     return pep_list, dataset, label


# def read_msp_with_decoy(msp_file):
#     """Read annotated spectra from msp file and return
#     data structure along with decoy peptides.
#     :param msp_file: str
#     :returns list
#     """
#
#     f = open(msp_file, "r")
#     lines = f.readlines()
#     f.close()
#
#     dataset = []
#     label = []
#
#     # FIXME: config should use only one get_config call.
#     spec_size = config.get_config(section='input', key='spec_size')
#     charge = config.get_config(section='input', key='charge')
#     use_mods = config.get_config(section='input', key='use_mods')
#
#     print('len of file: ' + str(len(lines)))
#     count = 0
#     limit = 200000
#     pep = 0
#     spec = []
#     is_name = is_mw = is_num_peaks = False
#     prev = 0
#     max_peaks = max_moz = 0
#     i = 0
#     while i < len(lines) and limit > 0:
#         line = lines[i]
#         i += 1
#         splits = line.split(':')
#         if (splits[0] == 'Name') and '_' in line:
#             split1 = splits[1]
#             l_charge = int(split1[split1.find('_') - 1])
#             if l_charge != charge and charge > 0:  # l_charge == l_charge always true.
#                 continue
#             if use_mods:
#                 pep = split1.split('/')[0].lstrip(' ')
#                 is_name = True
#             elif '(' not in splits[1] and ')' not in splits[1]:
#                 pep = split1.split('/')[0].lstrip(' ')
#                 is_name = True
#
#         if is_name and splits[0] == 'MW':
#             mass = float(splits[1])
#             if round(mass) < spec_size:
#                 is_mw = True
#                 # limit = limit - 1
#             else:
#                 is_name = is_mw = is_num_peaks = False
#                 continue
#
#         if is_name and is_mw and splits[0] == 'Num peaks':
#             num_peaks = int(splits[1])
#             if num_peaks > max_peaks:
#                 max_peaks = num_peaks
#
#             spec = np.zeros(spec_size)
#             while lines[i] != '\n':
#                 mz_line = lines[i]
#                 i += 1
#                 mz_splits = mz_line.split('\t')
#                 moz, intensity = float(mz_splits[0]), float(mz_splits[1])
#                 if moz > max_moz:
#                     max_moz = moz
#                 spec[round(moz)] += round(intensity)
#
#             spec = np.clip(spec, None, 1000.0)
#             spec = preprocessing.scale(spec)
#
#             is_num_peaks = True
#
#         if is_name and is_mw and is_num_peaks:
#             is_name = is_mw = is_num_peaks = False
#             # revPep = pep[0] + pep[1:-1][::-1] + pep[-1]
#             revPep = sim.get_rand_mod(pep)
#             if pep == revPep:
#                 print('decoy is the same. shuffling')
#                 # revPep = ''.join(rand.sample(revPep,len(revPep)))
#                 revPep = sim.get_rand_mod(pep, len(pep))
#                 print(pep)
#                 print(revPep)
#             t_spec = preprocessing.scale(sim.get_spectrum(pep))
#             rt_spec = preprocessing.scale(sim.get_spectrum(revPep))
#
#             # TODO: revert this back.
#             # dataset.append([spec, t_spec, rt_spec])
#             dataset.append([pep, spec, t_spec, rt_spec])
#             label.append([1, -1])
#
#             count = count + 1
#             pep = 0
#             spec = []
#             new = int((i / len(lines)) * 100)
#             if new > prev:
#                 clear_output(wait=True)
#                 print(str(new) + '%')
#                 prev = new
#
#     print('max peaks: ' + str(max_peaks))
#     print('count: ' + str(count))
#     print('max moz: ' + str(max_moz))
#     return dataset, label


def read_mgfs(folder_path):
    mgf_files = [f for f in listdir(folder_path) if isfile(join(folder_path, f)) and f.split('.')[-1] == 'mgf']
    assert len(mgf_files) > 0

    spec_size = config.get_config(section='input', key='spec_size')
    charge = config.get_config(section='input', key='charge')

    spectra = []
    masses = []
    charges = []
    for file in mgf_files:
        f = open(join(folder_path, file))
        spec_lines = f.readlines()
        f.close()

        if not spec_lines:
            continue

        spec = np.zeros(spec_size)
        isMass = False
        i = 0
        '''Read Headers'''
        while True:
            line = spec_lines[i]
            i += 1

            splits = line.split('=')
            if splits[0].upper() == 'PEPMASS':
                masses.append(float(splits[1].split(' ')[0]))
                isMass = True

            if isMass and splits[0].upper() == 'CHARGE':
                l_charge = int(splits[1][0])
                if charge and l_charge != charge:
                    del masses[-1]
                    isMass = False
                    isCharge = False
                else:
                    charges.append(l_charge)
                    isCharge = True
                break

        '''Read Spectrum'''
        while isMass and isCharge and i < len(spec_lines):
            line = spec_lines[i]
            i += 1

            if line != '\n' and 'END IONS' not in line.upper():
                splits = line.split(' ')
                moz, intensity = float(splits[0]), float(splits[1])
                spec[round(moz)] += round(intensity)
            elif 'END IONS' in line.upper():
                break

        if isMass and isCharge:
            spec = np.clip(spec, None, 1000.0)
            spec = preprocessing.scale(spec)
            spectra.append(spec)

    return spectra, masses, charges


def read_ms2(file):
    f = open(file)
    lines = f.readlines()
    f.close()

    spec_size = config.get_config(section='input', key='spec_size')
    charge = config.get_config(section='input', key='charge')

    spectra = []
    masses = []
    charges = []
    i = 0
    while i < len(lines):
        line = lines[i][:-1]
        i += 1

        splits = line.split('\t')
        if splits[0] == 'Z' and (charge <= 0 or float(splits[1]) == charge):
            charges.append(float(splits[1]))
            masses.append(float(splits[2]))
            spec = np.zeros(spec_size)
            while i < len(lines):
                line = lines[i][:-1]  # remove the \n character
                i += 1
                splits = line.split(' ')
                if 'S' in splits[0]:
                    break
                if 'Z' in splits[0]:
                    continue
                moz, intensity = float(splits[0]), float(splits[1])
                spec[round(moz)] += round(intensity)

            spec = np.clip(spec, None, 1000.0)
            spec = preprocessing.scale(spec)
            spectra.append(spec)

    return spectra, masses, charges



import torch
import torch.nn as nn
import torch.nn.functional as F

from specollate.src.snapconfig import config

#adding useless comment
class Net(nn.Module):
    def __init__(self, vocab_size, output_size=512, embedding_dim=512, hidden_lstm_dim=1024, lstm_layers=2):
        super(Net, self).__init__()
        self.spec_size = config.get_config(section='input', key='spec_size')
        self.seq_len = config.get_config(section='ml', key='pep_seq_len')
        self.output_size = output_size
        self.lstm_layers = lstm_layers
        self.hidden_lstm_dim = hidden_lstm_dim
        self.embedding_dim = embedding_dim
        self.vocab_size = vocab_size

        # dir_path = "/scratch/mtari008/job_" + os.environ['SLURM_JOB_ID'] + "/nist_massiv_80k_inch_aux-semi"
        # aux_means = np.load(join(dir_path, "aux_means.npy"))
        # aux_stds = np.load(join(dir_path, "aux_stds.npy"))
        # self.aux_means = torch.from_numpy(aux_means).float().to(0)
        # self.aux_stds = torch.from_numpy(aux_stds).float().to(0)
        
        ################### Spectra branch ###################
        # self.conv1_1    = nn.Conv1d(1, 32, 3, stride=1, padding=1)
        # self.maxpool1_1 = nn.MaxPool1d(50, stride=50)
        # self.linear1_1  = nn.Linear(int((self.spec_size*32)/50), 1024) # use with convo layers
        # self.linear1   = nn.Linear(35, 128)
        # self.linear2   = nn.Linear(128, 256)
        self.linear1_1 = nn.Linear(self.spec_size, 512)
        self.linear1_2 = nn.Linear(512, 256)
        # Spectra charge sub-branch
        # self.charge = config.get_config(section="input", key="charge")
        # self.linear_charge1 = nn.Linear(1024, self.charge)

        ################### Peptide branch ###################
        self.embedding = nn.Embedding(vocab_size, embedding_dim)
        self.lstm = nn.LSTM(embedding_dim, self.hidden_lstm_dim, self.lstm_layers,
                            # dropout=0.5,
                            batch_first=True, bidirectional=True)

        # self.linear2 = nn.Linear(111, 64)
        self.linear2_1 = nn.Linear(self.hidden_lstm_dim * 2, 512) # 2048, 1024
        self.linear2_2 = nn.Linear(512, 256)
        # NOTE: should the two layers below be just one layer.
        # self.linear_spec_mass1 = nn.Linear(1024, 256)
        # self.linear_spec_mass2 = nn.Linear(256, 1)
        # self.linear_spec_mass3 = nn.Linear(128, 1)
        # self.linear_pep_mass1 = nn.Linear(1024, 256)
        # self.linear_pep_mass2 = nn.Linear(256, 1)
        # self.linear_pep_mass3 = nn.Linear(128, 1)
        do = config.get_config(section="ml", key="dropout")
        self.dropout_conv1_1 = nn.Dropout(do)
        self.dropout1 = nn.Dropout(do)
        self.dropout2 = nn.Dropout(do)
        self.dropout1_1 = nn.Dropout(do)
        self.dropout1_2 = nn.Dropout(do)
        self.dropout1_3 = nn.Dropout(do)
        self.dropout1_4 = nn.Dropout(do)
        self.dropout1_5 = nn.Dropout(do)
        self.dropout2_1 = nn.Dropout(do)
        self.dropout2_2 = nn.Dropout(do)
        self.dropout2_3 = nn.Dropout(do)
        self.dropout2_4 = nn.Dropout(do)
        self.dropout2_5 = nn.Dropout(do)
        self.dropout2_6 = nn.Dropout(do)
        self.dropout_charge1 = nn.Dropout(do)
        self.dropout_spec_mass1 = nn.Dropout(do)
        self.dropout_spec_mass2 = nn.Dropout(do)
        self.dropout_spec_mass3 = nn.Dropout(do)
        self.dropout_pep_mass1 = nn.Dropout(do)
        self.dropout_pep_mass2 = nn.Dropout(do)
        self.dropout_pep_mass3 = nn.Dropout(do)
        print("dropout: {}".format(do))
        #self.dropout3 = nn.Dropout(0.3)
        
    def forward(self, data, data_type=None):
        assert not data_type or data_type == "specs" or data_type == "peps"
        res = []
        if not data_type or data_type == "specs":
            specs = data[0].squeeze()
            # mass_charge = specs[:, 0:35]
            # specs = specs[:, 35:]
            # out = F.relu(self.conv1_1(specs.view(-1, 1, self.spec_size)))
            # out = self.maxpool1_1(out)
            # out = self.dropout_conv1_1(out)

            # out = F.relu(self.linear1_1(out.view(out.size(0), -1))) # use this line with conv networks

            out = F.relu((self.linear1_1(specs.view(-1, self.spec_size))))
            out = self.dropout1_1(out)

            # mc_out = F.relu(self.linear1(mass_charge.view(-1, 35)))
            # mc_out = self.dropout1(mc_out)
            # mc_out = F.relu(self.linear2(mc_out))
            # mc_out = self.dropout2(mc_out)
            # out = torch.cat((mc_out, out), axis=1)

            out_spec = F.relu(self.linear1_2(out))
            # out_spec = out_spec * mc_out
            out_spec = F.normalize(out_spec)
            
            # out_ch = self.dropout_charge1(out)
            # out_ch = self.linear_charge1(out_ch)
            # out_ch = F.relu(out_ch)
            
            # out_mass = self.dropout_spec_mass1(out)
            # out_mass = self.linear_spec_mass1(out_mass)
            # out_mass = F.relu(out_mass)
            # out_mass = self.dropout_spec_mass2(out_mass)
            # out_mass = self.linear_spec_mass2(out_mass)
            # out_mass = F.relu(out_mass)
            # out_mass = self.dropout_spec_mass3(out_mass)
            # out_mass = self.linear_spec_mass3(out_mass)
            res.append(out_spec)
            # res.append(out_ch)
            # res.append(out_mass)
        if not data_type or data_type == "peps":
            for peps in data[1:3]:
                peps = peps.squeeze()
                # aux = peps[:, :111].float()
                # aux = (aux - self.aux_means) / self.aux_stds
                # peps = peps[:, 111:]
                embeds = self.embedding(peps)
                # embeds = self.one_hot_tensor(peps)
                hidden = self.init_hidden(len(peps))
                hidden = tuple([e.data for e in hidden])
                lstm_out, _ = self.lstm(embeds, hidden)
                lstm_out = lstm_out[:, -1, :]
                out = lstm_out.contiguous().view(-1, self.hidden_lstm_dim * 2)
                out = self.dropout2_1(out)

                out = F.relu((self.linear2_1(out)))
                out = self.dropout2_2(out)
                
                # m_out = F.relu(self.linear2(mass.view(-1, 27)))
                # m_out = self.dropout2(m_out)
                # out = torch.cat((m_out, out), axis=1)

                # aux_out = F.relu(self.linear2(aux))
                # aux_out = self.dropout2(aux_out)
                # out = torch.cat((aux_out, out), axis=1)

                out_pep = F.relu(self.linear2_2(out))
                out_pep = F.normalize(out_pep)
                
                # out_mass = self.dropout_pep_mass1(out)
                # out_mass = self.linear_pep_mass1(out_mass)
                # out_mass = F.relu(out_mass)
                # out_mass = self.dropout_pep_mass2(out_mass)
                # out_mass = self.linear_pep_mass2(out_mass)
                # out_mass = F.relu(out_mass)
                # out_mass = self.dropout_pep_mass3(out_mass)
                # out_mass = self.linear_pep_mass3(out_mass)
                res.append(out_pep)
                # res.append(out_mass)
        return res
    
    def init_hidden(self, batch_size):
        weight = next(self.parameters()).data
        hidden = (weight.new(self.lstm_layers * 2, batch_size, self.hidden_lstm_dim).zero_(),
                      weight.new(self.lstm_layers * 2, batch_size, self.hidden_lstm_dim).zero_())
        return hidden

    def one_hot_tensor(self, peps):
        batch_size = len(peps)
        src = torch.zeros((batch_size, self.seq_len), dtype=torch.float16, device='cuda')
        src[peps > 0] = 1.0
        one_hots = torch.zeros((batch_size, self.seq_len, self.vocab_size), dtype=torch.float16, 
                                device='cuda')
        one_hots.scatter_(2, peps.view(batch_size, self.seq_len, 1), src.view(batch_size, self.seq_len, 1))
        one_hots.requires_grad = True
        return one_hots
    
    def name(self):
        return "Net"


from torch.utils.data import Sampler
import random

from specollate.src.snapconfig import config


class PSMSampler(Sampler):
    r"""Samples labeled spectra randomly within a mega batch.

    Args:
        masses (Dataset): used to get appropriate indices.
    """
#     mega_batches: Sized

    def __init__(self, masses):
        self.batch_size = config.get_config(section="ml", key="batch_size")
        self.batch_queue = []
        
        self.masses = masses

        min_mass_mb1 = min(self.masses) + 0.2

        mb_size = 2 * self.batch_size
        self.mega_batches_1 = []
        self.mega_batches_2 = []
        prev = cutoff = 0
        # TODO: check for the end of the list and empty range
        for idx, mass in enumerate(self.masses):
            if idx < prev:
                continue
            if mass > min_mass_mb1:
                while cutoff <= idx:
                    cutoff = prev + mb_size
                while cutoff > len(self.masses):
                    cutoff -= self.batch_size
                if cutoff - prev == 0:
                    break
                self.mega_batches_1.append((prev, cutoff))
                prev = cutoff
                min_mass_mb1 = self.masses[cutoff - 1] + 0.2

        start = 0
        for mb1_range in self.mega_batches_1:
            end = int( ((mb1_range[1] - mb1_range[0]) / 2) + mb1_range[0])
            if mb1_range[1] - mb1_range[0] == self.batch_size:
                end = mb1_range[1]
            self.mega_batches_2.append((start, end))
            start = end
        if start < self.mega_batches_1[-1][1]:
            self.mega_batches_2.append((start, self.mega_batches_1[-1][1]))
        
        self.current_mb = []
        self.initial_len = int(self.mega_batches_1[-1][1])
        

    def __iter__(self):
        if not self.batch_queue:
            if self.current_mb is self.mega_batches_1:
                self.current_mb = self.mega_batches_2
            else:
                self.current_mb = self.mega_batches_1
            random.shuffle(self.current_mb)
            
            for mb in self.current_mb:
                mb_range = list(range(mb[0], mb[1]))
                random.shuffle(mb_range)
                assert len(mb_range) % self.batch_size == 0
                for i in range(0, len(mb_range), self.batch_size):
                    self.batch_queue.append(mb_range[i:i+self.batch_size])
            
        for batch in random.sample(self.batch_queue, len(self.batch_queue)):
            for idx in batch:
                yield idx
        self.batch_queue = []


    def __len__(self) -> int:
        return int(self.current_mb[-1][1]) if self.current_mb else self.initial_len


import os
import timeit
from os.path import join
import re
import pickle
import numpy as np

import torch
import torch.distributed as dist
import torch.multiprocessing as mp
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import BatchSampler
torch.manual_seed(1)

from .src.snapconfig import config
from .src.snaptrain import dataset, model, trainmodel, sampler
from .src.snaputils import simulatespectra as sim

# with redirect_output("deepSNAP_redirect.txtS"):
train_loss     = []
test_loss      = []
train_accuracy = []
test_accuracy  = []

def run_par(rank, world_size):
    model_name = config.get_config(section="ml", key="model_name") #first k is spec size second is batch size
    print("Training {}."
        .format(model_name))
    # model_name = "time-4096"
    # wandb.init(project="SpeCollate", entity="pcds")
    # wandb.run.name = "{}-{}".format(model_name, wandb.run.id)
    # wandb.config.learning_rate = 0.00005
    
    setup(rank, world_size)

    batch_size  = config.get_config(section="ml", key="batch_size")
    in_tensor_dir = config.get_config(section='preprocess', key='in_tensor_dir') # for raptor
    # in_tensor_dir = "/scratch/mtari008/job_" + os.environ['SLURM_JOB_ID'] + "/nist_massiv_80k_ch_graymass27-semi" # for comet
    
    train_peps, train_specs, train_masses, test_peps, test_specs, test_masses = read_split_listings(in_tensor_dir)

    train_sampler = sampler.PSMSampler(train_masses)
    test_sampler  = sampler.PSMSampler(test_masses)

    train_batch_sampler = BatchSampler(sampler=train_sampler, batch_size=batch_size, drop_last=False)
    test_batch_sampler = BatchSampler(sampler=test_sampler, batch_size=batch_size, drop_last=False)

    train_dataset = dataset.LabeledSpectra(in_tensor_dir, train_peps, train_specs)
    test_dataset  = dataset.LabeledSpectra(in_tensor_dir, test_peps, test_specs)

    vocab_size = train_dataset.vocab_size

    # train_sampler = torch.utils.data.distributed.DistributedSampler(
    #     train_dataset,
    #     num_replicas=world_size,
    #     rank=rank,
    #     shuffle=True
    # )
    # test_sampler = torch.utils.data.distributed.DistributedSampler(
    #     test_dataset,
    #     num_replicas=world_size,
    #     rank=rank
    # )

    train_loader = torch.utils.data.DataLoader(
        dataset=train_dataset, num_workers=8, collate_fn=psm_collate,
        # batch_sampler=train_batch_sampler
        batch_size=batch_size, shuffle=True
    )

    test_loader = torch.utils.data.DataLoader(
        dataset=test_dataset, num_workers=8, collate_fn=psm_collate,
        # batch_sampler=test_batch_sampler,
        batch_size=batch_size, shuffle=True
    )

    lr = config.get_config(section="ml", key="lr")
    num_epochs = config.get_config(section="ml", key="epochs")
    weight_decay = config.get_config(section="ml", key="weight_decay")
    margin = config.get_config(section="ml", key="margin")
    snp_weight = config.get_config(section="ml", key="snp_weight")
    ce_weight = config.get_config(section="ml", key="ce_weight")
    mse_weight = config.get_config(section="ml", key="mse_weight")

    if rank == 0:
        print("same as 0 with filter enabled.")
        print("Learning Rate: {}".format(lr))
        print("Weigh Decay: {}".format(weight_decay))
        print("snp weight: {}".format(snp_weight))
        print("ce weight: {}".format(ce_weight))
        print("mse weight: {}".format(mse_weight))
        print("margin: {}".format(margin))

    triplet_loss = nn.TripletMarginLoss(margin=margin, p=2, reduction="sum", swap=True)
    cross_entropy_loss = nn.CrossEntropyLoss(reduction="sum")
    mse_loss = nn.MSELoss(reduction="mean")
    # device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    if torch.cuda.is_available():
        torch.cuda.set_device(rank)

    model_ = model.Net(vocab_size, output_size=512, embedding_dim=512, hidden_lstm_dim=512, lstm_layers=2).to(rank)
    optimizer = optim.Adam(model_.parameters(), lr=lr, weight_decay=weight_decay)

    #model_ = torch.load("./models/hcd/single-mod-10-0.3-0.0005.pt")
    model_ = nn.parallel.DistributedDataParallel(model_, device_ids=[rank])
    # model_, optimizer = apex.amp.initialize(model_, optimizer, opt_level="O1")
    # model_ = apex.parallel.DistributedDataParallel(model_)
    # model_.load_state_dict(torch.load("models/hcd/512-embed-2-lstm-SnapLoss2D-80k-nist-massive-gmc-semi-r-10.pt")["model_state_dict"])
    #optimizer = optim.SGD(model_.parameters(), lr=lr)

    # Create directories to cache files on the scratch storage.
    # moved out of this function since only once per node is needed.
    # os.makedirs("/scratch/train_lstm/spectra/", 0o755, exist_ok=True)
    # os.makedirs("/scratch/train_lstm/peptides/", 0o755, exist_ok=True)

    # wandb.watch(model_)
    for epoch in range(num_epochs):
        l_epoch = (epoch * world_size) + rank
        print("Epoch: {}".format(l_epoch))
        # train_sampler.set_epoch(l_epoch)
        start_time = timeit.default_timer()
        loss = trainmodel.train(model_, rank, train_loader, triplet_loss, cross_entropy_loss, mse_loss, optimizer, l_epoch)
        trainmodel.test(model_, rank, test_loader, triplet_loss, cross_entropy_loss, mse_loss, l_epoch)
        elapsed = timeit.default_timer() - start_time
        print("time takes: {} secs.".format(elapsed))

        dist.barrier()

        if l_epoch % 1 == 0 and rank == 0:
            torch.save({
            'epoch': l_epoch,
            'model_state_dict': model_.state_dict(),
            'optimizer_state_dict': optimizer.state_dict(),
            'loss': loss,
            }, "./models/{}-{}.pt".format(model_name, l_epoch))
            # model_name = "single_mod-{}-{}.pt".format(epoch, lr)
            # print(wandb.run.dir)
            # torch.save(model_.state_dict(), join("./models/hcd/", model_name))
            # wandb.save("{}-{}.pt".format(model_name, l_epoch))
    
    cleanup()

def setup(rank, world_size):
    os.environ['MASTER_ADDR'] = 'localhost'
    os.environ['MASTER_PORT'] = str(config.get_config(section="input", key="master_port"))
    torch.cuda.set_device(rank)
    dist.init_process_group(backend='nccl', world_size=world_size, rank=rank)
    # dist.init_process_group(backend='nccl', world_size=world_size, rank=rank)

def cleanup():
    dist.destroy_process_group()


def apply_filter(filt, file_name):
    file_parts = []
    charge = 0
    mods = 0
    try:
        file_parts = re.search(r"(\d+)-(\d+)-(\d+.\d+)-(\d+)-(\d+).[pt|npy]", file_name)
        charge = int(file_parts[4])
        mods = int(file_parts[5])
    except:
        print(file_name)
        print(file_parts)
    
    if ((filt["charge"] == 0 or charge <= filt["charge"]) # change this back to <=
        and (mods <= filt["mods"])):
        return True
    
    return False


def psm_collate(batch):
    specs = torch.cat([item[0] for item in batch], 0)
    peps = torch.stack([item[1] for item in batch], 0)
    dpeps = torch.stack([item[2] for item in batch if item[2].nelement() > 0], 0)
    peps_set = set(map(tuple, peps.tolist()))
    dpeps_set = set(map(tuple, dpeps.tolist()))
    dpeps_list = list(dpeps_set - dpeps_set.intersection(peps_set))
    dpeps = torch.tensor(dpeps_list, dtype=torch.long)
    charges = torch.cat([item[3] for item in batch], 0)
    # spec_masses = torch.cat([item[4] for item in batch], 0).view(-1, 1)
    # pep_masses = torch.FloatTensor([item[5] for item in batch]).view(-1, 1)
    
    # aas = ['_PAD'] + list(config.AAMass.keys())
    # idx2aa = {i:a for i, a in enumerate(aas)}
    # dpep_masses = []
    # for dpep in dpeps:
    #     dpep = "".join([idx2aa[i.item()] for i in dpep if i > 0])
    #     dpep_masses.append(sim.get_pep_mass(dpep))
    # dpep_masses = torch.FloatTensor(dpep_masses).view(-1, 1)
    counts = np.array([item[4] for item in batch])
    # return [specs, peps, dpeps, charges, spec_masses, pep_masses, dpep_masses, counts]
    return [specs, peps, dpeps, charges, counts]

# drop_prob=0.5
# print(vocab_size)

def read_split_listings(l_in_tensor_dir):
    print(l_in_tensor_dir)

    print("Reading train test split listings from pickles.")
    with open(join(l_in_tensor_dir, "train_peps.pkl"), "rb") as trp:
        train_peps = pickle.load(trp)
    with open(join(l_in_tensor_dir, "train_specs.pkl"), "rb") as trs:
        train_specs = pickle.load(trs)
    with open(join(l_in_tensor_dir, "test_peps.pkl"), "rb") as tep:
        test_peps = pickle.load(tep)
    with open(join(l_in_tensor_dir, "test_specs.pkl"), "rb") as tes:
        test_specs = pickle.load(tes)

    train_masses = []
    for train_pep in train_peps:
        train_mass = float(re.search(r"(\d+)-(\d+.\d+).pep", train_pep)[2])
        train_masses.append(train_mass)

    test_masses = []
    for test_pep in test_peps:
        test_mass = float(re.search(r"(\d+)-(\d+.\d+).pep", test_pep)[2])
        test_masses.append(test_mass)

    train_peps, train_specs, train_masses = zip(*sorted(zip(train_peps, train_specs, train_masses), key=lambda x: x[2]))
    train_peps, train_specs, train_masses = list(train_peps), list(train_specs), list(train_masses)

    test_peps, test_specs, test_masses = zip(*sorted(zip(test_peps, test_specs, test_masses), key=lambda x: x[2]))
    test_peps, test_specs, test_masses = list(test_peps), list(test_specs), list(test_masses)

    return train_peps, train_specs, train_masses, test_peps, test_specs, test_masses


if __name__ == '__main__':

    # Initialize parser 
    # parser = argparse.ArgumentParser()
    
    # # Adding optional argument 
    # parser.add_argument("-j", "--job-id", help="No arguments should be passed. \
    #     Instead use the shell script provided with the code.") 
    # parser.add_argument("-p", "--path", help="Path to the config file.")
    # parser.add_argument("-s", "--server-name", help="Which server the code is running on. \
    #     Options: raptor, comet. Default: comet", default="comet")
    
    # # Read arguments from command line 
    # args = parser.parse_args() 
    
    # if args.job_id: 
    #     print("job_id: %s" % args.job_id)
    #     job_id = args.job_id

    # if args.path:
    #     print("job_id: %s" % args.path)
    #     scratch = args.path

    

    mp.freeze_support() # needed to package multiprocessing libraries
    mp.set_start_method('forkserver')
    config.PARAM_PATH = join((os.path.dirname(__file__)), "config.ini")
    
    do_learn = True
    save_frequency = 2
    
    # torch.manual_seed(0)
    # torch.cuda.manual_seed(0)

    num_gpus = torch.cuda.device_count()
    print("Num GPUs: {}".format(num_gpus))
    # mp.spawn(run_par, args=(num_gpus,), nprocs=num_gpus, join=True)
    run_par(0, 1)

    # model.linear1_1.weight.requires_grad = False
    # model.linear1_1.bias.requires_grad = False
    # model.linear1_2.weight.requires_grad = False
    # model.linear1_2.bias.requires_grad = False


import sys
import numpy as np
import torch
# import progressbar
from tqdm import tqdm

from specollate.src.snapconfig import config
from specollate.src.snaptrain import process


def ppm(val, ppm_val):
    return (ppm_val / 1000000) * val


def spec_collate(batch):
    specs = torch.stack([item for item in batch], 0)
    dummy_pep = np.zeros(config.get_config(section="ml", key="pep_seq_len"))
    dummy_pep = torch.from_numpy(dummy_pep).long().unsqueeze(0)
    return [specs, dummy_pep]


def pep_collate(batch):
    peps = torch.stack([item for item in batch], 0)
    dummy_spec = np.zeros(config.get_config(section="input", key="spec_size"))
    dummy_spec = torch.from_numpy(dummy_spec).float().unsqueeze(0)
    dummy_pep = np.zeros((2, config.get_config(section="ml", key="pep_seq_len") + 24))
    dummy_pep = torch.from_numpy(dummy_pep).long()#.unsqueeze(0)
    # tqdm.write("{}".format(peps.shape))
    # tqdm.write("{}".format(dummy_pep.shape))
    return [dummy_spec, peps, dummy_pep]


def get_search_mask(spec_masses, pep_masses, tol):
    l_tol = tol
    rows = []
    cols = []
    pep_min = pep_max = 0
    for row_id, spec_mass in enumerate(spec_masses):
        # min_mass = max(spec_mass - l_tol, 0.0)
        # max_mass = spec_mass + l_tol
        min_mass = max(spec_mass - ppm(spec_mass, l_tol), 0.0)
        max_mass = spec_mass + ppm(spec_mass, l_tol)
        while (pep_min < len(pep_masses) and 
               min_mass > pep_masses[pep_min]):
            pep_min += 1
        while (pep_max < len(pep_masses) and 
               max_mass > pep_masses[pep_max]):
            pep_max += 1
        # pep_min = max(pep_min - 1, 0)
        # pep_max = min(pep_max + 1, len(pep_masses) - 1)
            
        # if pep_max == pep_min:
        #     print(row_id, pep_max, pep_min)
        rows.extend([row_id] * (pep_max - pep_min))
        cols.extend(range(pep_min, pep_max))
    
    assert len(rows) == len(cols)
    mask = torch.zeros(len(spec_masses), len(pep_masses))
    mask[rows, cols] = 1
    return mask


def runModel(loader, s_model, in_type, device):
    with torch.no_grad():
        accurate_labels = 0
        all_labels = 0
        loss = 0
        out_out = torch.Tensor().cpu()
        pbar = tqdm(loader, file=sys.stdout)
        pbar.set_description('Running Model...')
        # with progressbar.ProgressBar(max_value=len(loader)) as bar:
        for batch in pbar:
            batch[0], batch[1] = batch[0].to(device), batch[1].to(device)
            if in_type == "specs":
                out_ = s_model(batch, data_type=in_type)[0]
            elif in_type == "peps":
                batch[2] = batch[2].to(device)
                out_ = s_model(batch, data_type=in_type)[0]
            out_out = torch.cat((out_out, out_.to("cpu")), dim=0)
            # bar.update(batch_idx)
        return out_out


def search(search_loader, datasets, embeddings, device):
    pep_sort_inds = []
    pep_sort_vals = []
    dec_sort_inds = []
    dec_sort_vals = []

    spec_dataset = datasets["spec_dataset"]
    pep_dataset  = datasets["pep_dataset"]
    dec_dataset  = datasets["dec_dataset"]

    e_peps = embeddings["e_peps"]
    e_decs = embeddings["e_decs"]

    keep_psms = config.get_config(key="keep_psms", section="search")
    precursor_tolerance = config.get_config(key="precursor_tolerance", section="search")
    tol_type = config.get_config(key="precursor_tolerance_type", section="search")

    pbar = tqdm(search_loader, file=sys.stdout)
    pbar.set_description('Running Database Search...')
    # with progressbar.ProgressBar(max_value=len(search_loader)) as bar:
    for idx, spec_batch in enumerate(pbar):
        l_tol = precursor_tolerance
        batch_size = search_loader.batch_size
        st = idx * batch_size
        en = st + batch_size
        spec_masses = spec_dataset.spec_mass_list[st:en]
        
        if tol_type == "ppm":
            min_mass = max(spec_masses[0] - ppm(spec_masses[0], l_tol), 0)
            max_mass = spec_masses[-1] + ppm(spec_masses[-1], l_tol)
        else:
            min_mass = max(spec_masses[0] - l_tol, 0)
            max_mass = spec_masses[-1] + l_tol

        pep_min = pep_max = 0

        while (pep_min < len(pep_dataset.pep_mass_list) and 
                min_mass - pep_dataset.pep_mass_list[pep_min] > 0.01):
            pep_min += 1
        while (pep_max < len(pep_dataset.pep_mass_list) and
                max_mass - pep_dataset.pep_mass_list[pep_max] >= 0.01):
            pep_max += 1

        dec_min = dec_max = 0
        while (dec_min < len(dec_dataset.pep_mass_list) and
                min_mass - dec_dataset.pep_mass_list[dec_min] > 0.01):
            dec_min += 1
        while (dec_max < len(dec_dataset.pep_mass_list) and 
                max_mass - dec_dataset.pep_mass_list[dec_max] >= 0.01):
            dec_max += 1

        pep_batch = e_peps[pep_min:pep_max]
        pep_masses = pep_dataset.pep_mass_list[pep_min:pep_max]
        dec_batch = e_decs[dec_min:dec_max]
        dec_masses = dec_dataset.pep_mass_list[dec_min:dec_max]

        spec_batch = spec_batch.to(device)
        #print("pep batch len: {}".format(len(pep_batch)))
        l_pep_batch_size = 16384
        # l_pep_batch_size = 32768
        pep_loader = torch.utils.data.DataLoader(
            dataset=pep_batch, batch_size=l_pep_batch_size)
        l_pep_dist = []
        for pep_idx, l_pep_batch in enumerate(pep_loader):
            l_pep_batch = l_pep_batch.to(device)
            l_st = pep_idx * l_pep_batch_size
            l_en = l_st + l_pep_batch_size
            l_pep_masses = pep_masses[l_st:l_en]
            spec_pep_mask = get_search_mask(spec_masses, l_pep_masses, precursor_tolerance).to(device)
            # spec_pep_mask[spec_pep_mask == 0] = float("inf")
            spec_pep_dist = 1.0 / process.pairwise_distances(spec_batch, l_pep_batch)
            l_pep_dist.append((spec_pep_dist * spec_pep_mask).to("cpu"))
        
        l_pep_dist.append(torch.zeros(len(spec_batch), keep_psms + 1))
        pep_sort = torch.cat(l_pep_dist, 1)
        pep_lcn = np.ma.masked_array(pep_sort, mask=pep_sort==0).min(1).data
        pep_sort = pep_sort.sort(descending=True)
        pep_sort_inds.append(pep_sort.indices[:, :keep_psms + 1] + pep_min) # offset for the global array
        pep_sort_vals.append(torch.cat((pep_sort.values[:, :keep_psms + 1], 
                                        torch.from_numpy(pep_lcn).unsqueeze(1)), 1))
        
        dec_loader = torch.utils.data.DataLoader(
            dataset=dec_batch, batch_size=l_pep_batch_size)
        l_dec_dist = []
        for dec_idx, l_dec_batch in enumerate(dec_loader):
            l_dec_batch = l_dec_batch.to(device)
            l_st = dec_idx * l_pep_batch_size
            l_en = l_st + l_pep_batch_size
            l_dec_masses = dec_masses[l_st:l_en]
            spec_dec_mask = get_search_mask(spec_masses, l_dec_masses, precursor_tolerance).to(device)
            # spec_dec_mask[spec_dec_mask == 0] = float("inf")
            spec_dec_dist = 1.0 / process.pairwise_distances(spec_batch, l_dec_batch)
            l_dec_dist.append((spec_dec_dist * spec_dec_mask).to("cpu"))
        
        l_dec_dist.append(torch.zeros(len(spec_batch), keep_psms + 1))
        dec_sort = torch.cat(l_dec_dist, 1)
        dec_lcn = np.ma.masked_array(dec_sort, mask=dec_sort==0).min(1).data
        dec_sort = dec_sort.sort(descending=True)
        dec_sort_inds.append(dec_sort.indices[:, :keep_psms + 1] + dec_min) # offset for the global array
        dec_sort_vals.append(torch.cat((dec_sort.values[:, :keep_psms + 1], 
                                        torch.from_numpy(dec_lcn).unsqueeze(1)), 1))
        # bar.update(idx)
    
    pep_inds = torch.cat(pep_sort_inds, 0)
    pep_vals = torch.cat(pep_sort_vals, 0)
    dec_inds = torch.cat(dec_sort_inds, 0)
    dec_vals = torch.cat(dec_sort_vals, 0)
    return pep_inds, pep_vals, dec_inds, dec_vals


import pickle
import re
import time
import random as rand
from os.path import join

from sklearn.model_selection import train_test_split

from .src.snapconfig import config

def apply_filter(l_filt, file_name):
    try:
        file_parts = re.search(r"(\d+)-(\d+)-(\d+.\d+)-(\d+)-(\d+).[pt|npy]", file_name)
        l_charge = int(file_parts[4])
        mods = int(file_parts[5])
    except:
        print(file_name)
        print(file_parts)
    
    if ((l_filt["charge"] == 0 or l_charge <= l_filt["charge"]) # change this back to <=
        and (mods <= l_filt["mods"])):
        return True
    
    return True#False

def load_file_names(l_filt, l_listing_path, count=None):
    'Load the peptide and corresponding spectra file names that satisfy the filter'
    with open(l_listing_path, 'rb') as f:
        dir_listing = pickle.load(f)

    rand.shuffle(dir_listing)
    l_pep_file_names = []
    l_spec_file_names_lists = []
    for pep, spec_list in dir_listing[:count]:
        spec_file_list = []
        for spec in spec_list:
            if apply_filter(l_filt, spec):
                spec_file_list.append(spec)
        if spec_file_list:
            l_pep_file_names.append(pep)
            l_spec_file_names_lists.append(spec_file_list)

    assert len(l_pep_file_names) == len(l_spec_file_names_lists)
    return l_pep_file_names, l_spec_file_names_lists

if __name__ == '__main__':
    charge      = config.get_config(section='input', key='charge')
    use_mods    = config.get_config(section='input', key='use_mods')
    num_mods    = config.get_config(section='input', key='num_mods')
    filt        = {'charge': charge, 'mods': num_mods if use_mods else 0}
    test_size   = config.get_config(section='ml', key='test_size')
    train_count = config.get_config(section="ml", key="train_count")
    batch_size  = config.get_config(section="ml", key="batch_size")
    train_count = None if train_count == 0 else train_count

    in_tensor_dir = config.get_config(section='preprocess', key='in_tensor_dir')
    print(in_tensor_dir)
    listing_path = join(in_tensor_dir, 'pep_spec.pkl')
    pep_file_names, spec_file_names_lists = load_file_names(filt, listing_path, train_count)
    
    split_rand_state = int(time.time())
    train_peps, test_peps, train_specs, test_specs = train_test_split(
        pep_file_names, spec_file_names_lists, test_size=test_size,
        random_state=split_rand_state, shuffle=True)
    
    # test_peps, val_peps, test_specs, val_specs = train_test_split(
    #     test_peps, test_specs, test_size=.1,
    #     random_state=split_rand_state, shuffle=True)
    # get the 100k version
    # train_peps  = train_peps[:80000]
    # train_specs = train_specs[:80000]
    # test_peps   = test_peps[:20000]
    # test_specs  = test_specs[:20000]

    print("Writing train test split listings as pickles.")
    with open(join(in_tensor_dir, "train_peps.pkl"), "wb") as trp:
        pickle.dump(train_peps, trp)
    with open(join(in_tensor_dir, "train_specs.pkl"), "wb") as trs:
        pickle.dump(train_specs, trs)
    with open(join(in_tensor_dir, "test_peps.pkl"), "wb") as tep:
        pickle.dump(test_peps, tep)
    with open(join(in_tensor_dir, "test_specs.pkl"), "wb") as tes:
        pickle.dump(test_specs, tes)
    # with open(join(in_tensor_dir, "val_peps.pkl"), "wb") as vap:
    #     pickle.dump(test_peps, vap)
    # with open(join(in_tensor_dir, "val_specs.pkl"), "wb") as vas:
    #     pickle.dump(test_specs, vas)



import numpy as np
import torch

from specollate.src.snapconfig import config
from specollate.src.snaputils import simulatespectra as sim


def pairwise_distances(x, y=None):
    """
    Input: x is a Nxd matrix
           y is an optional Mxd matirx
    Output: dist is a NxM matrix where dist[i,j] is the square norm between x[i,:] and y[j,:]
            if y is not given then use 'y=x'.
    i.e. dist[i,j] = ||x[i,:]-y[j,:]||^2
    """
    x_norm = (x ** 2).sum(1).view(-1, 1)
    if y is not None:
        y_t = torch.transpose(y, 0, 1)
        y_norm = (y ** 2).sum(1).view(1, -1)
    else:
        y_t = torch.transpose(x, 0, 1)
        y_norm = x_norm.view(1, -1)

    dist = x_norm + y_norm - 2.0 * torch.mm(x, y_t)
    # Ensure diagonal is zero if x=y
    if y is None:
        #dist = dist - torch.diag(dist.diag())
        dist.fill_diagonal_(0.0)
    dist[torch.isnan(dist)] = 0.0  # set all nan values to zero
    return torch.clamp(dist, 0.0, np.inf)


def process_fasta_in_batch(model, file_path, spectra_batch_size):
    f = open(file_path)
    lines = f.readlines()
    f.close()

    batch_size = config.get_config(section='ml', key='batch_size')

    masses = []
    spectra_out = []
    peps = []

    start = 0
    i = 0
    while start < len(lines):
        print('Batch: ' + str(i))
        i += 1

        print('Generating spectra...')
        spectra, l_masses, l_peps = sim.fasta_to_spectra(lines, start, spectra_batch_size, dh)
        masses.extend(l_masses)
        peps.extend(l_peps)
        start = start + spectra_batch_size

        with torch.no_grad():
            print('Converting to tensor...')
            '''dtype=torch.float'''
            spectra = np.asarray(spectra)
            spectraTensor = torch.as_tensor(spectra, dtype=torch.float)[:, None, :]
            spectra_loader = torch.utils.data.DataLoader(dataset=spectraTensor, batch_size=batch_size, shuffle=False)

            print('Running the model...')
            spectra_out.extend(run_model(model, spectra_loader))

    return spectra_out, masses, peps


def run_model(model, loader):
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    with torch.no_grad():
        out = []
        for batch_idx, data in enumerate(loader):
            data = data.to(device)
            out.extend(model(data)[0].cpu().detach().numpy())
    print(len(out))
    return out



import re
import random as rand
from heapq import merge
import numpy as np

from sklearn import preprocessing

from specollate.src.snapconfig import config


def get_rand_mod(seq, num_mods=1):
    """
    Get num_mods number of random modifications added to peptide seq.
    :param seq: str
    :param num_mods: int
    :return: str
    """
    aas = list(config.AAMass.keys())
    res = temp = seq
    for i in range(num_mods):
        while res == temp:
            rand_indx = rand.randint(0, len(seq) - 1)
            rand_mod = aas[rand.randint(0, len(aas)) - 1]
            temp = temp[:rand_indx] + rand_mod + temp[rand_indx + 1:]
        res = temp
    return res


def gray_code(num):
    return num ^ (num >> 1)


def decimal_to_binary_array(num, arr_len):
    bin_arr = [float(i) for i in list('{0:0b}'.format(num))]
    assert len(bin_arr) <= arr_len
    res = [0] * (arr_len - len(bin_arr)) + bin_arr
    return res


def get_aa_mass(aa):
    """
    Get amino acid mass from the given aa character.
    :param aa: char
    :return: float
    """
    return config.AAMass[aa] + 57.021464 if aa == 'C' else config.AAMass[aa]

def get_mod_aa_mass(aa):
    """
    Get amino acid mass from the given (modified/unmodified) aa.
    :param aa: char/s
    :return: float
    """
    return sum(config.AAMass[sub_aa] for sub_aa in aa)


def get_pep_mass(pep):
    """
    Get peptide mass from the given pep string.
    :param pep: str
    :return: float
    """
    return sum(config.AAMass[aa] for aa in pep) + config.H2O


def get_spectrum(seq):
    """
    Get theoretical spectrum from a peptide string seq.
    :param seq: str
    :return: int[]
    """

    spec_size = config.get_config(section='input', key='spec_size')
    # charge = config.get_config(section='input', key='charge')

    if len(seq) == 0:
        print('Error: seq length is zero.')
        return []
    
    first = ""
    if seq[0].islower():
        first = seq[0]
        seq = seq[1:]
    
    pep_parts = re.findall(r"([A-Z][a-z]?)", seq)
    pep_parts[0] = first + pep_parts[0]

    b_spectrum = []
    y_spectrum = []

    b_spectrum.append(get_aa_mass(seq[0]) + config.PROTON)
    y_spectrum.append(get_aa_mass(seq[-1]) + config.H2O + config.PROTON)

    for i, (faa, baa) in enumerate(zip((seq[1:]), seq[-2::-1])):
        b_spectrum.append(b_spectrum[i] + get_aa_mass(faa))
        y_spectrum.append(y_spectrum[i] + get_aa_mass(baa))

    merged_out = list(merge(b_spectrum, y_spectrum))
    if merged_out[-1] > spec_size:
        print('Error: peptide mass {} is larger than {}'.format(merged_out[-1], spec_size))
        print(seq)
    t_spec = np.zeros(spec_size)
    t_spec[np.rint(merged_out).astype(int)] = 1
    return t_spec


def get_mod_spectrum(seq):
    """
    Get theoretical spectrum from a peptide string seq.
    :param seq: str
    :return: int[]
    """

    spec_size = config.get_config(section='input', key='spec_size')
    # charge = config.get_config(section='input', key='charge')

    if len(seq) == 0:
        print('Error: seq length is zero.')
        return []
    
    first = ""
    if seq[0].islower():
        first = seq[0]
        seq = seq[1:]
    
    pep_parts = re.findall(r"([A-Z][a-z]?)", seq)
    pep_parts[0] = first + pep_parts[0]

    b_spectrum = []
    y_spectrum = []

    b_spectrum.append(get_mod_aa_mass(pep_parts[0]) + config.PROTON)
    y_spectrum.append(get_mod_aa_mass(pep_parts[-1]) + config.H2O + config.PROTON)

    for i, (faa, baa) in enumerate(zip((seq[1:]), pep_parts[-2::-1])):
        b_spectrum.append(b_spectrum[i] + get_mod_aa_mass(faa))
        y_spectrum.append(y_spectrum[i] + get_mod_aa_mass(baa))

    merged_out = list(merge(b_spectrum, y_spectrum))
    if merged_out[-1] > spec_size:
        print('Error: peptide mass {} is larger than {}'.format(merged_out[-1], spec_size))
        print(seq)
        
    return merged_out


def get_mod_spectrum_hyperscore(seq):
    """
    Get theoretical spectrum from a peptide string seq.
    :param seq: str
    :return: int[]
    """

    if len(seq) == 0:
        print('Error: seq length is zero.')
        return []
    
    first = ""
    if seq[0].islower():
        first = seq[0]
        seq = seq[1:]
    
    pep_parts = re.findall(r"([A-Z][a-z]?)", seq)
    pep_parts[0] = first + pep_parts[0]

    b_spectrum = []
    y_spectrum = []

    b_spectrum.append(get_mod_aa_mass(pep_parts[0]) + config.PROTON)
    y_spectrum.append(get_mod_aa_mass(pep_parts[-1]) + config.H2O + config.PROTON)

    for i, (faa, baa) in enumerate(zip((seq[1:]), pep_parts[-2::-1])):
        b_spectrum.append(b_spectrum[i] + get_mod_aa_mass(faa))
        y_spectrum.append(y_spectrum[i] + get_mod_aa_mass(baa))
        
    return b_spectrum, y_spectrum


def fasta_to_spectra(lines, start, count, dh):
    t_spectra = []
    masses = []
    peps = []

    prev = 0
    end = min(start + count, len(lines))
    for i, line in enumerate(lines[start:end]):
        splits = line.split('\t')

        pep = splits[0]
        # print(pep)
        peps.append(pep)
        spec = get_spectrum(pep)
        t_spectra.append(preprocessing.scale(spec))
        masses.append(float(splits[1]))

        # print(splits[1])
        """Progress Monitor"""
        new = int(((i + start) / len(lines)) * 100)
        if new > prev:
            dh.update(str(new) + '%')
            prev = new

    return t_spectra, masses, peps
